package be.iminds.ilabt.jfed.lowlevel.api.test;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.*;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AMv2Wrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AMv3Wrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.testing.base.ApiTest;
import be.iminds.ilabt.jfed.util.TargetAuthority;

import javax.inject.Inject;
import java.net.URL;
import java.util.*;

/**
 * SimpleTestAggregateManager2. One simple test of "Geni Aggregate Manager API v2"
 * <p/>
 * This is little more than a connectivity test and one basic test of XmlRpc and GetVersion.
 * <p/>
 * Use TestAggregateManager2 for a lot more tests of the "Geni Aggregate Manager API v2". (And also for more thorough tests of GetVersion)
 */
public class TestAnyServiceGetVersion extends ApiTest {
    private CommonAMTest commonAMTest;

    @Inject
    public TestAnyServiceGetVersion(Logger logger, TargetAuthority testedAuthority, GeniUser user, JFedConnectionProvider connectionProvider, AuthorityListModel authorityListModel, JFedPreferences jFedPreferences) {
        super(logger, testedAuthority, user, connectionProvider, authorityListModel, jFedPreferences);
    }

    public String getTestDescription() {
        return "Calls GetVersion on any available Aggregate manager for the given testbed. This is mainly useful to test connectivity.";
    }

    @Override
    public List<String> getOptConfKeys() {
        List<String> res = new ArrayList<String>();

        res.add("service_name");
        res.add("service_version");

        CommonAMTest.addOptionsForConnectionSshProxy(res);

        return res;
    }

    /**
     * Check for correctness of a AM2 XmlRpc result. Should be tested for each reply received.
     */
    public void testAM2CorrectnessXmlRpcResult(Hashtable res) {
        Object code = res.get("code");
        Object value = res.get("value");
        Object output = res.get("output");
        assertNotNull(code, "testAM2CorrectnessXmlRpcResult code == null in " + res);
        assertNotNull(value, "testAM2CorrectnessXmlRpcResult value == null in " + res);
        assertEquals(code.getClass(), Hashtable.class, "testAM2CorrectnessXmlRpcResult code is not Hashtable in " + res + " (it is " + code.getClass().getName() + " with value \"" + code.toString() + "\")");

        Hashtable codeHt = (Hashtable) code;
        Object genicode = codeHt.get("geni_code");
        assertNotNull(genicode, "testAM2CorrectnessXmlRpcResult code does not contain \"geni_code\" in " + res);
        assertEquals(genicode.getClass(), Integer.class, "testAM2CorrectnessXmlRpcResult code.geni_code is not int in " + res);
        int genicodevalue = (Integer) genicode;

        //output should be present if code is not 0
        if (genicodevalue != 0) {
            assertNotNull(output, "testAM2CorrectnessXmlRpcResult: while geni_code is non success (" + genicodevalue + "), output == null in " + res);
            assertEquals(output.getClass(), String.class, "testAM2CorrectnessXmlRpcResult: while geni_code is non success (" + genicodevalue + "), output is not String (it is " + output.getClass().getName() + " with value \"" + output.toString() + "\") in \"+res+\"");
        } else {
            //this is a connectivity test, we don't warn here, we just make a note of it.
            if (output == null)
                note("testAM2CorrectnessXmlRpcResult: while geni_code is success (" + genicodevalue + "), output is ommited. This is allowed by the API.");
            else {
                if (output != null && !(output instanceof String))
                    note("testAM2CorrectnessXmlRpcResult: while geni_code is success (" + genicodevalue + "), output is not String (it is " + output.getClass().getName() + " with value \"" + output.toString() + "\"). This is allowed but not recommended by jFed.");
            }
        }
    }

    private URL preferredUrl = null;
    private ServerType preferredServerType = null;
    public void setUp() {
        String serviceName = getTestConfig().getProperty("service_name");
        String serviceVersion = getTestConfig().getProperty("service_version");
        if (serviceName != null)
            note("Will limit to service: "+serviceName);
        if (serviceVersion != null)
            note("Will limit to service version: "+serviceVersion);

        SfaAuthority auth = testedAuthority.getSfaAuthorityToConnect();
        Map<ServerType, URL> urls = auth.getUrls();

        if (urls.isEmpty()) {
            fatalError("No URLs are known for the tested authority "+auth.getUrnString());
        } else
            note("The tested authority has "+urls.size()+" registered URLs");

        for (Iterator<Map.Entry<ServerType, URL>> it = urls.entrySet().iterator(); it.hasNext(); ) {
            Map.Entry<ServerType, URL> e = it.next();
            if (serviceName != null && !e.getKey().getRole().name().equalsIgnoreCase(serviceName)) {
                it.remove(); continue;
            }
            if (serviceVersion != null && !e.getKey().getVersion().equalsIgnoreCase(serviceVersion)) {
                it.remove(); continue;
            }
        }

        if (urls.isEmpty()) {
            fatalError("Due to the service_name and/or service_version filter, no URLs are left to test.");
        }

        List<ServerType> preferences = Arrays.asList(
                new ServerType(ServerType.GeniServerRole.AM, 3),
                new ServerType(ServerType.GeniServerRole.AM, 2),
                new ServerType(ServerType.GeniServerRole.PROTOGENI_SA, 1),
                new ServerType(ServerType.GeniServerRole.GENI_CH_SA, 2),
                new ServerType(ServerType.GeniServerRole.GENI_CH_SA, 1),
                new ServerType(ServerType.GeniServerRole.GENI_CH_MA, 2),
                new ServerType(ServerType.GeniServerRole.GENI_CH_MA, 1),
                new ServerType(ServerType.GeniServerRole.GENI_CH, 2),
                new ServerType(ServerType.GeniServerRole.GENI_CH, 1),
                new ServerType(ServerType.GeniServerRole.PlanetLabSliceRegistry, 1),
                new ServerType(ServerType.GeniServerRole.PROTOGENI_CH, 1),
                new ServerType(ServerType.GeniServerRole.SCS, 1)
        );

        for (ServerType pref : preferences) {
            URL url = urls.get(pref);
            if (url != null) {
                preferredUrl = url;
                preferredServerType = pref;
                break;
            }
        }
        if (preferredUrl == null) {
            Map.Entry<ServerType, URL> e = urls.entrySet().iterator().next();
            preferredUrl = e.getValue();
            preferredServerType = e.getKey();
        }

        note("Chosen servertype: "+preferredServerType);
        note("URL: "+preferredUrl);

        commonAMTest = new CommonAMTest(this, user.getUserAuthority());
        commonAMTest.setupProxyForConnectionPool(getTestConfig());

        note("done");
    }

    private Hashtable versionRawResult = null;

    @Test
    public void testGetVersion() throws JFedException {
        switch (preferredServerType.getRole()) {
            case AM: { testAMGetVersion(); break; }
            case GENI_CH:    //fall through
            case GENI_CH_MA: //fall through
            case GENI_CH_SA: { testFedApiGetVersion(); break; }
            case SCS: { testScsGetVersion(); break; }
            default: { fatalError("There is no known \"GetVersion\" test for the server type "+preferredServerType); break; }
        }
    }

    public void testScsGetVersion() throws JFedException {
        StitchingComputationService scs = new StitchingComputationService(getLogger(), getJFedPreferences());

        SfaConnection sfaCon = (SfaConnection) getConnectionProvider().getConnectionByAuthority(getUser(), getTestedAuthority().getSfaAuthorityToConnect(), preferredServerType);
        StitchingComputationService.SCSReply<Hashtable> reply = scs.getVersion(sfaCon);
        Hashtable versionRawResult = reply.getRawResult();

        assertNotNull(versionRawResult);

        assertEquals(reply.getCode(), GeniAMResponseCode.GENIRESPONSE_SUCCESS.getCode(), "GeniResponse code is not SUCCESS (0)");

        Hashtable<String, Object> replyHt = AbstractApi.apiSpecifiesHashtableStringToObject(versionRawResult);
        Hashtable<String, Object> valueHt = AbstractApi.apiSpecifiesHashtableStringToObject(replyHt.get("value"));
        Object iface = valueHt.get("interface");
        assertEquals(iface, "scs");
//        Object versions = valueHt.get("geni_api_versions");
//        assertNotNull(versions);

        note("Test completed successfully.");
    }

    public void testFedApiGetVersion() throws JFedException {
        FederationSliceAuthorityApi2 sa = new FederationSliceAuthorityApi2(getLogger(), getJFedPreferences());

        SfaConnection sfaCon = (SfaConnection) getConnectionProvider().getConnectionByAuthority(getUser(), getTestedAuthority().getSfaAuthorityToConnect(), preferredServerType);
        AbstractFederationApi.FederationApiReply<FederationSliceAuthorityApi2.GetVersionSAResult> reply = sa.getVersion(sfaCon);
        Hashtable versionRawResult = reply.getRawResult();

        assertNotNull(versionRawResult);

        assertEquals(reply.getCode(), GeniAMResponseCode.GENIRESPONSE_SUCCESS.getCode(), "GeniResponse code is not SUCCESS (0)");

        Hashtable<String, Object> replyHt = AbstractApi.apiSpecifiesHashtableStringToObject(versionRawResult);
        Hashtable<String, Object> valueHt = AbstractApi.apiSpecifiesHashtableStringToObject(replyHt.get("value"));
        Object version = valueHt.get("VERSION");
        assertNotNull(version);

        note("Test completed successfully. Version is "+version+"");
    }

    public void testAMGetVersion() throws JFedException {
        AggregateManagerWrapper amWrapper = getAutomaticAggregateManagerWrapperFactory().create(logger, testedAuthority.getSfaAuthorityToConnect());
        versionRawResult = amWrapper.getVersion();

        testAM2CorrectnessXmlRpcResult(versionRawResult);

        assertNotNull(versionRawResult);

        Hashtable<String, Object> replyHt = AbstractApi.apiSpecifiesHashtableStringToObject(versionRawResult);
        Hashtable<String, Object> codeHt = AbstractApi.apiSpecifiesHashtableStringToObject(replyHt.get("code"));
        Integer genicodeHt = AbstractApi.apiSpecifiesInt(codeHt.get("geni_code"));

        assertEquals(genicodeHt, GeniAMResponseCode.GENIRESPONSE_SUCCESS.getCode(), "GeniResponse code is not SUCCESS (0)");

        Hashtable<String, Object> valueHt = AbstractApi.apiSpecifiesHashtableStringToObject(replyHt.get("value"));
        Integer geniapiHt = AbstractApi.apiSpecifiesInt(replyHt.get("geni_api"));

        assertNotNull(geniapiHt);

        note("Test completed successfully. Version is"+geniapiHt+".\nAll Done\n");
    }
}
