package be.iminds.ilabt.jfed.ssh_terminal_tool.putty;

import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpFile;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.*;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

@Singleton
public class PageantHelper {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(PageantHelper.class);

    private final JFedPreferences jFedPreferences;
    private final LinkedBlockingQueue<SshKeyInfo> registerKeyQueue = new LinkedBlockingQueue<>();
    private final Set<SshKeyInfo> keyHistory = Collections.synchronizedSet(new HashSet<>());
    private boolean startedPageantBackgroundProcess = false;
    private Thread pageantThread;

    @Inject
    public PageantHelper(JFedPreferences jFedPreferences) {
        this.jFedPreferences = jFedPreferences;
    }

    /**
     * Non-blocking call. Registers key with pageant.
     *
     * @param sshKeyInfo null is also allowed (does nothing then)
     */
    public synchronized void registerKey(@Nullable final SshKeyInfo sshKeyInfo) {
        if (sshKeyInfo == null) return;

        registerKeys(Collections.singletonList(sshKeyInfo));
    }

    public synchronized void registerKeys(List<SshKeyInfo> sshKeyInfos) {
        if (sshKeyInfos == null || sshKeyInfos.isEmpty()) return;

        sshKeyInfos = new ArrayList<>(sshKeyInfos); //copy this to a new arraylist to be sure that we can do operations on it!

        if (sshKeyInfos.removeIf(keyHistory::contains)) {
            LOG.info("One or more keys were already registered with pageant, and are therefore ignored.");
        }

        if (sshKeyInfos.isEmpty()) return;

        registerKeyQueue.addAll(sshKeyInfos);

        if (pageantThread == null)
            startPageantThread();
    }

    private synchronized void startPageantThread() {
        if (pageantThread != null)
            throw new RuntimeException("Pageant Thread has already started! Cannot run more than 1 instance...");

        LOG.debug("Starting Pageant thread.");
        pageantThread = new Thread(new PageantKeyAdder());
        pageantThread.setName("Pageant-Thread");
        pageantThread.setDaemon(true); //=> this application may exit before putty stops
        pageantThread.start();
        LOG.debug("Started Pageant thread.");
    }


    private class PageantKeyAdder implements Runnable {
        @Override
        public void run() {
            while (!Thread.interrupted()) {
                try {
                    //Fetch new keys to be added.
                    final List<SshKeyInfo> sshKeyInfos = tryFetchSshKeyInfos();

                    if (sshKeyInfos.isEmpty())
                        continue;

                    final List<TmpFile> tmpFiles = new ArrayList<>();
                    try {
                        final LinkedList<String> command = new LinkedList<>();

                        /*
                        Add the files, starting with unprotected, and ending with the ones that ask passphrase.
                        This is a good order, because the unprotected ones will likely be removed soon,
                        while the protected ones are typically not removed.
                        However, it still leaves some chance that a file gets deleted before pageant.exe deals with it.
                        */
                        sshKeyInfos.stream()
                                .filter(sshKeyInfo -> sshKeyInfo != null && sshKeyInfo.getPuttyKeyFile() != null)
                                .forEach(sshKeyInfo -> {
                                    TmpFile tmpUnlockedKeyFile = sshKeyInfo.getUnlockedPuttyKeyFile();
                                    if (tmpUnlockedKeyFile != null) {
                                        tmpFiles.add(tmpUnlockedKeyFile);
                                        tmpUnlockedKeyFile.store();
                                        command.addFirst(tmpUnlockedKeyFile.getFilename()); //creates the file if needed. (release later deletes it again)
                                    } else {
                                        //jFed doesn't have key to PPK file. User will have to provide pageant with it.
                                        TmpFile tmpKeyFile = sshKeyInfo.getPuttyKeyFile();
                                        if (tmpKeyFile != null) {
                                            tmpFiles.add(tmpKeyFile);
                                            tmpKeyFile.store();
                                            command.addLast(tmpKeyFile.getFilename()); //creates the file if needed. (release later deletes it again)
                                        }
                                    }
                                });

                        command.addFirst(jFedPreferences.getFile(GuiPreferenceKey.PREF_PUTTY_DIRECTORY).getPath() + File.separator + "pageant.exe");

                        if (!startedPageantBackgroundProcess) {
                            registerKeysInPageantWithBackgroundProcess(sshKeyInfos, command);

                            startedPageantBackgroundProcess = true;
                            try {
                                Thread.sleep(5000); //give pageant some time to start, otherwise, we might delete key files to early
                            } catch (InterruptedException ignored) {
                            }

                        } else {
                            registerKeysInPageant(sshKeyInfos, command);

                            try {
                                Thread.sleep(1000); //wait a while before allowing next pageant to be started
                            } catch (InterruptedException ignored) {
                            }
                        }
                    } catch (IOException e) {
                        LOG.error("IOException while trying to launch Pageant: {}", e.getMessage(), e);
                    } finally {
                        LOG.debug("removing unencrypted Pageant key files");
                        //remove private key files (if needed)
                        tmpFiles.forEach(TmpFile::delete);
                    }
                } catch (InterruptedException ignored) {
                    //occurs only when take() was interrupted
                }
            }
            LOG.debug("Pageant thread was interrupted and returned.");
            pageantThread = null;
        }

        private void registerKeysInPageantWithBackgroundProcess(List<SshKeyInfo> sshKeyInfos, List<String> command) {
            LOG.debug("Starting Pageant first time, so starting in background thread. Not waiting for it to stop, as it might never.");
            Thread backgroundPageantThread = new Thread(() -> {
                //just launch the first command to add a key
                try {
                    LOG.debug("Starting first pageant.exe in background thread");
                    executePageant(command);

                    //key is registered, let it be known
                    keyHistory.addAll(sshKeyInfos);
                } catch (IOException e) {
                    LOG.error("First pageant.exe threw IOException", e);
                }
            });
            backgroundPageantThread.setName("FirstPageant");
            backgroundPageantThread.setDaemon(true);
            backgroundPageantThread.start();
        }

        private void registerKeysInPageant(List<SshKeyInfo> sshKeyInfos, List<String> command) throws IOException {
            LOG.debug("Starting Pageant and waiting for it to stop");
            executePageant(command);

            //key is registered, let it be known
            keyHistory.addAll(sshKeyInfos);
        }

        private void executePageant(List<String> command) throws IOException {
            LOG.info("Pageant command: " + command);
            Process p = Runtime.getRuntime().exec(command.toArray(new String[command.size()]));
            p.getOutputStream().close();
            BufferedReader input =
                    new BufferedReader(new InputStreamReader(p.getInputStream()));
            String line;
            while ((line = input.readLine()) != null) {
                LOG.debug("pageant.exe output: " + line);
            }
            input.close();
            LOG.debug("Pageant has no more output");
            try {
                p.waitFor();
            } catch (InterruptedException ignored) {
            }
            LOG.debug("Pageant returned");
        }

        /**
         * Returns a list of keys to be added
         * <p>
         * Because keys typically arrive in batches, we try to wait for them to arrive before processing them.
         *
         * @return List with new SSH Keys to be added into pageant
         * @throws InterruptedException
         */
        private List<SshKeyInfo> tryFetchSshKeyInfos() throws InterruptedException {
            final List<SshKeyInfo> sshKeyInfos = new ArrayList<>();
            final SshKeyInfo firstSshKeyInfo =
                    registerKeyQueue.take(); //blocking call, waits until item in queue
            sshKeyInfos.add(firstSshKeyInfo);

            //wait at least 100ms for additional keys.
            boolean othersAvailable = true;
            while (othersAvailable) {
                final SshKeyInfo additionalSshKeyInfo = registerKeyQueue.poll(100, TimeUnit.MILLISECONDS);
                if (additionalSshKeyInfo != null) {
                    sshKeyInfos.add(additionalSshKeyInfo);
                } else {
                    othersAvailable = false;
                }
            }
            return sshKeyInfos;
        }
    }


}
