package be.iminds.ilabt.jfed.experimenter_gui.controller.model;

import be.iminds.ilabt.jfed.experimenter_gui.util.ProxyServiceUtil;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.rspec.model.LoginService;
import be.iminds.ilabt.jfed.rspec.model.controller.ExperimentCommand;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import be.iminds.ilabt.jfed.util.*;
import ch.ethz.ssh2.Connection;
import ch.ethz.ssh2.ConnectionInfo;
import javafx.application.Platform;
import javafx.beans.property.*;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableBooleanValue;
import javafx.beans.value.ObservableObjectValue;
import javafx.beans.value.ObservableValue;
import javafx.collections.FXCollections;
import javafx.concurrent.Task;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import javax.security.sasl.AuthenticationException;
import java.net.InetSocketAddress;
import java.net.SocketImplFactory;
import java.net.SocketTimeoutException;
import java.util.Base64;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * SshConnectionPool looks at the FXRspecModel and provides access to connections and to properties reflecting
 * if connections to nodes are available or can be made available
 */
public class SshConnectionPool implements ExperimentCommand.SshConnectionProvider {
    private final static Logger LOG = LoggerFactory.getLogger(SshConnectionPool.class);

    private final ProxyPreferencesManager proxyPreferencesManager;
    private final ProxyServiceUtil proxyServiceUtil;

    //Mapping nodeID's to their SSH-connection and other info
    private final Map<String, NodeInfo> nodeInfoByUniqueId = new HashMap<>();

    private final ObservableObjectValue<FXModelRspec> modelRspecValue;

    private final ReadOnlyListWrapper<String> nodeIds =
            new ReadOnlyListWrapper<>(FXCollections.<String>observableArrayList());

    private final BooleanProperty allConnectable = new SimpleBooleanProperty(false);
    private final BooleanProperty allConnected = new SimpleBooleanProperty(false);
    private final BooleanProperty anyConnectable = new SimpleBooleanProperty(false);
    private final BooleanProperty anyConnected = new SimpleBooleanProperty(false);

    private final GeniUser sshUser;

    public SshConnectionPool(ObservableObjectValue<FXModelRspec> modelRspecValue, GeniUser sshUser,
                             ProxyPreferencesManager proxyPreferencesManager,
                             ProxyServiceUtil proxyServiceUtil) {
        this.modelRspecValue = modelRspecValue;
        this.sshUser = sshUser;
        this.proxyPreferencesManager = proxyPreferencesManager;
        this.proxyServiceUtil = proxyServiceUtil;

        modelRspecValue.addListener(new ChangeListener<FXModelRspec>() {
            @Override
            public void changed(ObservableValue<? extends FXModelRspec> observableValue, FXModelRspec old, FXModelRspec nw) {
                updateModel();
            }
        });
        updateModel();

//        //visual debugging
//        if (LOG.isDebugEnabled()) {
//            SshConnectionPoolDebugPanel sshConnectionPoolDebugPanel = new SshConnectionPoolDebugPanel(this);
//            sshConnectionPoolDebugPanel.show();
//        }
    }

    private ChangeListener<NodeState> updateAllConnectionStateListener = new ChangeListener<NodeState>() {
        @Override
        public void changed(ObservableValue<? extends NodeState> observableValue, NodeState nodeState, NodeState nodeState2) {
            updateAllConnectionState();
        }
    };

    private void updateAllConnectionState() {
        boolean newAllConnectable = true;
        boolean newAllConnected = true;
        boolean newAnyConnectable = false;
        boolean newAnyConnected = false;
        for (NodeInfo nodeInfo : nodeInfoByUniqueId.values()) {
            switch (nodeInfo.state.get()) {
                case NO_INFO: {
                    newAllConnectable = false;
                    newAllConnected = false;
                    break;
                }
                case CONNECTABLE: {
                    newAnyConnectable = true;
                    newAllConnected = false;
                    break;
                }
                case CONNECTING: {
                    newAnyConnectable = true;
                    newAllConnected = false;
                    break;
                }
                case CONNECTED: {
                    newAnyConnectable = true;
                    newAnyConnected = true;
                    break;
                }
                case FAILCONNECT: {
                    newAllConnected = false;
                    break;
                }
            }
//            if (!newAllConnectable && !newAllConnected &&
//                    newAllConnectable && newAnyConnected)
//                break;
        }
        allConnectable.setValue(newAllConnectable);
        allConnected.setValue(newAllConnected);
        anyConnectable.setValue(newAnyConnectable);
        anyConnected.setValue(newAnyConnected);
    }

    private void updateModel() {
        if (modelRspecValue.get() == null)
            return;

        for (FXRspecNode n : modelRspecValue.get().getNodes()) {
            String nodeId = n.getUniqueId();
            getNodeInfo(nodeId).update(n);
        }
    }

    private void closeAll() {
        for (NodeInfo ni : nodeInfoByUniqueId.values()) {
            ni.close();
        }
    }

    public ReadOnlyListProperty<String> getNodeIds() {
        return nodeIds.getReadOnlyProperty();
    }

    private NodeInfo getNodeInfo(String nodeId) {
        NodeInfo ni = nodeInfoByUniqueId.get(nodeId);
        if (ni == null) {
            ni = new NodeInfo(nodeId);
            ni.state.addListener(updateAllConnectionStateListener);
            nodeInfoByUniqueId.put(nodeId, ni);
            nodeIds.get().add(nodeId);
        }
        assert ni.nodeId.equals(nodeId);
        return ni;
    }

    public ObservableObjectValue<NodeState> getNodeState(String nodeId) {
        return getNodeInfo(nodeId).state;
    }

    public ObservableBooleanValue isNodeConnectable(String nodeId) {
        return getNodeInfo(nodeId).state.isEqualTo(NodeState.NO_INFO).not();
    }

    public ObservableBooleanValue isNodeConnected(String nodeId) {
        return getNodeInfo(nodeId).state.isEqualTo(NodeState.CONNECTED);
    }

    public ObservableBooleanValue allConnectableProperty() {
        return allConnectable;
    }

    public ObservableBooleanValue allConnectedProperty() {
        return allConnected;
    }

    public BooleanProperty anyConnectableProperty() {
        return anyConnectable;
    }

    public BooleanProperty anyConnectedProperty() {
        return anyConnected;
    }

    /**
     * @param nodeId       node to get connection to
     * @param blockingMode
     * @return
     */
    public Connection getConnection(String nodeId, ExperimentCommand.BlockingMode blockingMode) {
        NodeInfo ni = getNodeInfo(nodeId);
        switch (blockingMode) {
            case NON_BLOCKING: {
                switch (ni.state.get()) {
                    case NO_INFO:
                    case FAILCONNECT:
                        return null;
                    case CONNECTABLE: {
                        ni.startConnect();
                        return null;
                    }
                    case CONNECTING:
                        return null;
                    case CONNECTED:
                        return ni.connection;
                    default:
                        throw new RuntimeException("unsupported case " + ni.state.get());
                }
            }
            case WAIT_FOR_SINGLE: {
                switch (ni.state.get()) {
                    case NO_INFO:
                    case FAILCONNECT:
                        return null;
                    case CONNECTABLE: {
                        ni.startConnect();
                        ni.waitForConnect();
                        return ni.connection;
                    }
                    case CONNECTING: {
                        ni.waitForConnect();
                        return ni.connection;
                    }
                    case CONNECTED:
                        return ni.connection;
                    default:
                        throw new RuntimeException("unsupported case " + ni.state.get());
                }
            }
            case WAIT_FOR_ALL: {
                throw new RuntimeException("WAIT_FOR_ALL not yet implemented");
            }
            default:
                throw new RuntimeException("unsupported blockingMode " + blockingMode);
        }
    }


    /**
     * Creates an SSH-connection for every node. Every connection is
     * authenticated with the info provided by loginServices.
     * <p>
     * The connections are initialised in another thread, so this call is non-blocking.
     */
    public void initAllConnection() {
        for (FXRspecNode node : modelRspecValue.get().getNodes()) {
            final String nodeID = node.getUniqueId();
            NodeInfo nodeInfo = getNodeInfo(nodeID);
            NodeState state = nodeInfo.state.get();
            if (!state.equals(NodeState.NO_INFO) && !state.equals(NodeState.CONNECTED))
                nodeInfo.startConnect();
        }
    }

    @Nullable
    private LoginService chooseLoginService(@Nullable List<LoginService> loginServices) {
        if (loginServices == null || loginServices.isEmpty())
            return null;

        LoginService chosenLoginService = loginServices.get(0);
        for (int i = 1; i < loginServices.size(); i++) {
            LoginService otherLoginService = loginServices.get(i);
            if (otherLoginService.getUsername().equals(sshUser.getUserUrn().getEncodedResourceName()))
                chosenLoginService = otherLoginService;
        }

        if (chosenLoginService == null || chosenLoginService.getHostname() == null || chosenLoginService.getIntPort() <= 0) {
            LOG.error(" invalid chosenLoginService skipped: " + chosenLoginService);
            return null;
        }

        return chosenLoginService;
    }

    private class NodeInfo {
        final String nodeId;

        Connection connection;
        Thread connectionCreatorThread;

        ObjectProperty<NodeState> state = new SimpleObjectProperty<NodeState>(NodeState.NO_INFO);

        private LoginService loginService;
        private JFedConnection.SshProxyInfo proxyService;

        private NodeInfo(String nodeId) {
            this.nodeId = nodeId;
        }

        public void update(FXRspecNode rspecNode) {
            List<LoginService> loginServices = rspecNode.getLoginServices();

            if (loginServices != null && !loginServices.isEmpty()) {
                loginService = chooseLoginService(rspecNode.getLoginServices());
                if (loginService != null)
                    proxyService = proxyServiceUtil.findTestbedProxy(rspecNode, loginService);
            } else {
                loginService = null;
                proxyService = null;
            }

            if (loginService == null) {
                close();
                state.setValue(NodeState.NO_INFO);
            } else {
                if (state.get().equals(NodeState.NO_INFO))
                    state.set(NodeState.CONNECTABLE);
            }
        }

        public void close() {
            if (connection != null) {
                connection.close();
                connection = null;
            }
            if (loginService != null)
                state.set(NodeState.CONNECTABLE);
            else
                state.setValue(NodeState.NO_INFO);
        }

        public void startConnect() {
            if (state.equals(NodeState.NO_INFO))
                return;

            Task<Void> task = new Task<Void>() {
                @Override
                protected Void call() throws Exception {
                    blockingConnect();
                    return null;
                }
            };

            //The thread will automatically die when the runnable returns
            connectionCreatorThread = new Thread(task);
            connectionCreatorThread.start();
        }

        public void waitForConnect() {
            if (connectionCreatorThread != null)
                try {
                    connectionCreatorThread.join();
                    connectionCreatorThread = null;
                } catch (InterruptedException e) {
                    LOG.debug("connectionCreatorThread.join() interrupted", e);
                }
        }

        public Connection blockingConnect() {
            int attemptCount = 1;
            boolean success = false;
            safeSetState(NodeState.CONNECTING);
            //Sometimes the connection gave a timeout on the first try, so retry once if this is the case
            while (attemptCount < 3 && !success) {
                try {
                    UserSshKeyInfo userSshKeyInfo = new UserSshKeyInfo(sshUser);

                    SshKeyInfo sshKeyInfo;

                    //Dit gaf omwille van locked keys problemen
                    //ISSUE: https://ibcn-jira.intec.ugent.be/browse/FEDIBBTDEV-406

                    //                        SshKeyInfo overrideSshKeyInfo = PreferencesUtil.getOverriddenSshKeyInfo();
                    //                        if (overrideSshKeyInfo != null) {
                    //                            LOG.info(
                    //                                    "ExternalSshTerminal: The preferences are overriding the SSH key used(" + overrideSshKeyInfo.getClass().getName() + ").");
                    //                            sshKeyInfo = overrideSshKeyInfo;
                    //                        } else {
                    //                            LOG.info("ExternalSshTerminal: The user's SSH key will be used.");
                    sshKeyInfo = userSshKeyInfo;
                    //                        }

                    //LOG.info("ControllerModel: Public key = " + KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(sshKeyInfo.getPublicKey()));
                    //LOG.info("ControllerModel: Private key length = " + (sshKeyInfo.getPrivateKey() == null
                    //        ? "null" : "" + sshKeyInfo.getPrivateKey().getEncoded().length));

                    SocketImplFactory socketImplFactory = null;

                    JFedConnection.SshProxyInfo sshProxyInfo =
                            (JFedConnection.SshProxyInfo)
                                    proxyPreferencesManager.getSshProxySettings(sshUser.getUserAuthority().getProxies(),
                                            sshUser, userSshKeyInfo);

                    LOG.debug("Creating SSH connection to " + loginService.getHostname() + ":" + loginService.getIntPort());
                    if (sshProxyInfo != null) {

                        SshServerProxyHelper.SshProxyInfo sshProxyInfo2 = new SshServerProxyHelper.SshProxyInfo(
                                new InetSocketAddress(sshProxyInfo.getHostname(), sshProxyInfo.getPort()),
                                sshProxyInfo.getUsername(),
                                sshProxyInfo.getHostKey(),
                                new String(KeyUtil.privateKeyToAnyPem(sshProxyInfo.getSshKeyInfo().getPrivateKey()))
                        );

                        socketImplFactory
                                = GanymedSshSocketImplFactory.getFactory(sshProxyInfo2);
                        assert socketImplFactory != null;

                        LOG.debug("Using proxy for connection");
                    }

                    if(proxyService != null){
                        LOG.info("Wrapping SSH connection into proxy defined by testbed: {}", proxyService);

                        SshServerProxyHelper.SshProxyInfo proxyServiceSshProxyInfo = new SshServerProxyHelper.SshProxyInfo(
                                new InetSocketAddress(proxyService.getHostname(), proxyService.getPort()),
                                proxyService.getUsername(),
                                proxyService.getHostKey(),
                                new String(KeyUtil.privateKeyToAnyPem(proxyService.getSshKeyInfo().getPrivateKey()))
                        );

                        socketImplFactory = GanymedSshSocketImplFactory.getFactory(proxyServiceSshProxyInfo, socketImplFactory);
                    }


                    connection = new Connection(socketImplFactory,
                            loginService.getHostname(),
                            loginService.getIntPort());

                    ConnectionInfo conInf = connection.connect();
                    if (conInf != null)
                        LOG.debug("Connected to server. Server hostkey is: " + Base64.getEncoder().encodeToString(conInf.serverHostKey));
                    else {
                        LOG.debug("Connection to server seems to have failed (ConnectionInfo is null), will try to continue anyway.");
                    }

                    SSHKeyHelper sshkeyHelper =
                            new SSHKeyHelper(sshKeyInfo.getPublicKey(), sshKeyInfo.getPrivateKey());

                    boolean isAuthenticated =
                            connection.authenticateWithPublicKey(loginService.getUsername(), sshkeyHelper.getPEMAnyPrivateKey(), "nopass");

                    if (!isAuthenticated) {
                        String[] remainingMethodsArray;
                        try {
                            remainingMethodsArray = (connection == null) ? new String[]{}
                                    : connection.getRemainingAuthMethods(loginService.getUsername());
                        } catch (Exception | AssertionError e) {
                            LOG.error("Exception while calling getRemainingAuthMethods: " + e.getMessage(), e);
                            remainingMethodsArray = new String[]{};
                        }
                        String remainingMethods = "";
                        for (String remainingMethod : remainingMethodsArray)
                            remainingMethods += " " + remainingMethod;
                        if (remainingMethodsArray.length == 0)
                            remainingMethods = "   no other authentication methods returned by server.";
                        else
                            remainingMethods = "   Remaining authentication methods:" + remainingMethods + " ";

                        LOG.warn("Could not authenticate connection for FXRspecNode: " + nodeId + ", for username: " + loginService.getUsername() + remainingMethods);
                        throw new AuthenticationException("Could not authenticate connection for FXRspecNode: " + nodeId + ", for username: " + loginService.getUsername() + remainingMethods);
                    }

                    success = true;
                    safeSetState(NodeState.CONNECTED);
                } catch (AssertionError ex) {
                    LOG.error("Could not setup a connection to node {} due to AssertionError", nodeId, ex);
                    safeSetState(NodeState.FAILCONNECT);
                } catch (Exception ex) {
                    attemptCount++;
                    if (attemptCount < 3 && ex instanceof SocketTimeoutException) {
                        try {
                            Thread.sleep(100);
                        } catch (InterruptedException e) {
                            LOG.debug("InterruptedException while waiting for next connect attempt. Will cancel connecting.", e);
                            safeSetState(NodeState.FAILCONNECT);
                            return null;
                        }
                        LOG.info("Connection for FXRspecNode " + nodeId + " gave a SocketTimeoutException. Retrying once more..");
                    }
                    //Log if not a SocketTimeoutException or if it still occured after 2 tries
                    else {
                        LOG.error("Could not setup a connection to node {}", nodeId, ex);
                        safeSetState(NodeState.FAILCONNECT);
                    }
                }
            }

            return connection;
        }

        public void safeSetState(final NodeState newState) {
            Platform.runLater(new Runnable() {
                @Override
                public void run() {
                    state.setValue(newState);
                }
            });
        }

    }

    public enum NodeState {NO_INFO, CONNECTABLE/*but unconnected*/, CONNECTING, CONNECTED, FAILCONNECT}
}
