package be.iminds.ilabt.jfed.experimenter_gui.editor.addresspool_properties;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecEditor;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.CMIManager;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.ComponentManagerInfo;
import be.iminds.ilabt.jfed.experimenter_gui.health.HealthStatus;
import be.iminds.ilabt.jfed.experimenter_gui.util.StringConverters;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXAddressPool;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import javafx.beans.InvalidationListener;
import javafx.beans.Observable;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.ComboBox;
import javafx.scene.control.Label;
import javafx.scene.control.TextField;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.stage.Stage;
import org.controlsfx.validation.ValidationSupport;
import org.controlsfx.validation.Validator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Comparator;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 4/23/2015
 * Time: 16:46
 */
public class AddressPoolPropertiesDialog extends BorderPane {
    private static final Logger LOG = LoggerFactory.getLogger(AddressPoolPropertiesDialog.class);

    private final FXAddressPool addressPool;
    private final FXModelRspec modelRspec;

    private final JFedConfiguration config;
    private final AuthorityList authorityList;
    private final CMIManager cmiManager;

    private final ValidationSupport validationSupport;

    private final ObjectProperty<Integer> currentlyAvailable = new SimpleObjectProperty<>(null);


    private final ObjectProperty<ComponentManagerInfo.Resources.RefreshAdvertisementTask> activeRefreshTask =
            new SimpleObjectProperty<>(null);

    @FXML
    private TextField nameTextField;
    @FXML
    private ComboBox<ComponentManagerInfo> componentManagerInfoComboBox;
    @FXML
    private TextField countTextField;

    @FXML
    private HBox ipFreeHBox;
    @FXML
    private Label ipFreeLabel;

    @FXML
    private Button saveButton;
    @FXML
    private Button cancelButton;
    private boolean success = false;

    AddressPoolPropertiesDialog(FXAddressPool addressPool, FXModelRspec modelRspec, JFedConfiguration config,
                                AuthorityList authorityList, CMIManager cmiManager) {
        this.addressPool = addressPool;
        this.modelRspec = modelRspec;

        this.config = config;
        this.authorityList = authorityList;
        this.cmiManager = cmiManager;


        //initialize the Scene
        FXMLUtil.injectFXML(this);

        componentManagerInfoComboBox.getSelectionModel().selectedItemProperty().addListener(new InvalidationListener() {
            @Override
            public void invalidated(Observable observable) {
                ComponentManagerInfo cmi = componentManagerInfoComboBox.getSelectionModel().getSelectedItem();

                JFedConfiguration.TestbedDescription testbedDescription = config.getTestbedDescription(cmi.getAuthority().getUrn());

                if(testbedDescription == null){
                    LOG.warn("Could not find testbed description!");
                    return;
                }

                final HealthStatus healthStatus = testbedDescription.getHealthStatus();
                if(healthStatus != null
                        && healthStatus.getIpv4sTotal() != null
                        && healthStatus.getIpv4sTotal() > 0){
                    ipFreeLabel.setText(healthStatus.getIpv4sAvailable() + " of " + healthStatus.getIpv4sTotal());
                    ipFreeHBox.setVisible(true);

                }   else{
                    ipFreeHBox.setVisible(false);
                }

            }
        });


        config.getTestbedDescriptions().stream().filter(JFedConfiguration.TestbedDescription::hasAddressPoolSupport)
                .map(JFedConfiguration.TestbedDescription::getUrn)
                .map(cmiManager::getComponentManagerInfo)
                .collect(Collectors.toCollection(componentManagerInfoComboBox::getItems));

        componentManagerInfoComboBox.getItems().sort(Comparator.comparing(cmi -> cmi.getAuthority().getHrn().toLowerCase()));
        componentManagerInfoComboBox.setConverter(StringConverters.CMI_STRING_CONVERTER);


        //load values from the provided channel
        nameTextField.setText(addressPool.getClientId());

        if (addressPool.getComponentManagerId() != null) {
            ComponentManagerInfo cmi = cmiManager.getComponentManagerInfo(addressPool.getComponentManagerId());
            if (cmi != null)
                componentManagerInfoComboBox.getSelectionModel().select(cmi);
            else
                LOG.warn("Could not find CMI with URN {}", addressPool.getComponentManagerId());
        } else {
            componentManagerInfoComboBox.getSelectionModel().clearSelection();
        }
        countTextField.setText(addressPool.getCount() != null ? Integer.toString(addressPool.getCount()) : "0");


        //initialize validationSupport
        validationSupport = new ValidationSupport();
        validationSupport.registerValidator(nameTextField, Validator.createPredicateValidator(new Predicate<String>() {
            @Override
            public boolean test(String s) {
                return ModelRspecEditor.isValidNodeName(s);
            }
        }, "You must provide a client ID"));
        validationSupport.registerValidator(componentManagerInfoComboBox, Validator.createEmptyValidator("You must choose a testbed"));
        validationSupport.registerValidator(countTextField, Validator.createPredicateValidator(new Predicate<String>() {
            @Override
            public boolean test(String s) {
                try {
                    return Integer.parseInt(s) > 0;
                } catch (NumberFormatException ex) {
                    return false;
                }
            }
        }, "You must provide a positive number of IPs to request"));

        saveButton.disableProperty().bind(validationSupport.invalidProperty());
    }


    @FXML
    public void onSaveAction(ActionEvent actionEvent) {
        addressPool.setClientId(nameTextField.getText());
        addressPool.setComponentManagerId(componentManagerInfoComboBox.getSelectionModel().getSelectedItem().getAuthority().getUrn());
        addressPool.setCount(Integer.parseInt(countTextField.getText()));

        success = true;

        Stage stage = (Stage) saveButton.getScene().getWindow();
        stage.close();

    }

    public boolean isSuccess() {
        return success;
    }

    @FXML
    public void onCancelAction(ActionEvent actionEvent) {
        Stage stage = (Stage) cancelButton.getScene().getWindow();
        stage.close();
    }

}
