package be.iminds.ilabt.jfed.experimenter_gui.editor.ontology;

import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.controller.NodePropertyOverview;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.model.query.SparqlService;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.webmap.ClusteringWebMap;
import be.iminds.ilabt.jfed.experimenter_gui.util.ui.StageUtils;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.collections.transformation.FilteredList;
import javafx.fxml.FXML;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.ComboBox;
import javafx.scene.layout.BorderPane;
import javafx.stage.Stage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * User: twalcari
 * Date: 1/8/2015
 * Time: 14:34
 */
public class ResultMapDialog extends BorderPane implements ClusteringWebMap.ClusteringWebMapListener {
    private static final Logger LOG = LoggerFactory.getLogger(ResultMapDialog.class);
    @FXML
    private ComboBox<SparqlService.NodeInfo> nodesComboBox;

    @FXML
    private Button selectNodeButton;

    @FXML
    private ClusteringWebMap webMap;

    private final ObservableList<SparqlService.NodeInfo> nodes = FXCollections.observableArrayList();
    private final FilteredList<SparqlService.NodeInfo> nodesWithinBounds =
            new FilteredList<SparqlService.NodeInfo>(nodes);

    private Map<SparqlService.NodeInfo, ComparableLatLong> nodeLocations = new HashMap<>();


    private ObjectProperty<SparqlService.NodeInfo> result = new SimpleObjectProperty<>(null);
//    private Marker selectedLocationMarker;

    //dialog variables
    private final Stage dialogStage;

    public ResultMapDialog() {
        FXMLUtil.injectFXML(this);


        nodesComboBox.setConverter(NodePropertyOverview.NODE_INFO_STRING_CONVERTER);
        selectNodeButton.disableProperty().bind(nodesComboBox.valueProperty().isNull());

        Scene scene = new Scene(this);

        dialogStage = new Stage();
        StageUtils.setJFedStageIcons(dialogStage);
        dialogStage.setScene(scene);

        dialogStage.setTitle("Map of query results");
        dialogStage.setResizable(true);
    }


    @Override
    public void onMarkerClicked(String id) {
        selectNode(nodes.stream().filter(node -> node.getComponentUrn().toString().equals(id)).findFirst().orElse(null));
    }

    @Override
    public void onBoundsChanged(double minLat, double maxLat, double minLon, double maxLon) {
        nodesWithinBounds.setPredicate(node ->
                minLat >= node.getLongitude()
                        && maxLat <= node.getLatitude()
                        && minLon >= node.getLongitude()
                        && maxLon <= node.getLongitude());
    }

    private static final class ComparableLatLong {
        private final double lat, lon;

        public ComparableLatLong(double lat, double lon) {
            this.lon = lon;
            this.lat = lat;
        }

        public double getLat() {
            return lat;
        }

        public double getLon() {
            return lon;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ComparableLatLong that = (ComparableLatLong) o;

            if (Double.compare(that.lat, lat) != 0) return false;
            if (Double.compare(that.lon, lon) != 0) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result;
            long temp;
            temp = Double.doubleToLongBits(lat);
            result = (int) (temp ^ (temp >>> 32));
            temp = Double.doubleToLongBits(lon);
            result = 31 * result + (int) (temp ^ (temp >>> 32));
            return result;
        }
    }


    private void addNodesToMap() {
        Set<ComparableLatLong> uniqueLocations = new HashSet<>();

        nodeLocations.clear();
        webMap.clearMarkers();

        //first setup locations, but try to prevent duplicate locations by shifting them a bit if necessary
        nodes.stream().filter(node -> node.getLatitude() != null && node.getLongitude() != null)
                .forEach(node -> {

                            ComparableLatLong ll = new ComparableLatLong(node.getLatitude(), node.getLongitude());

                            boolean changeLat = true;
                            while (uniqueLocations.contains(ll)) {
                                if (changeLat)
                                    ll = new ComparableLatLong(ll.getLat() + 0.0001, ll.getLon());
                                else
                                    ll = new ComparableLatLong(ll.getLat(), ll.getLon() + 0.0001);
                                changeLat = !changeLat;
                            }
                            uniqueLocations.add(ll);

                            nodeLocations.put(node, ll);
                        }


                );

        for (Map.Entry<SparqlService.NodeInfo, ComparableLatLong> entry : nodeLocations.entrySet()) {
            String id = entry.getKey().getComponentUrn().toString();

            webMap.addMarker(id, entry.getValue().getLat(), entry.getValue().getLon());
        }
    }


    private void clearNodeSelection() {
        nodesComboBox.setValue(null);
    }

    private void selectNode(SparqlService.NodeInfo node) {
        assert (node.getLatitude() != null && node.getLongitude() != null);
        //map.panTo(nodeLocations.get(node));

        nodesComboBox.setValue(node);
    }

    public void setNodes(List<SparqlService.NodeInfo> nodes) {
        this.nodes.setAll(FXCollections.observableList(nodes));
        nodesComboBox.setItems(nodesWithinBounds);

        addNodesToMap();
    }

    @FXML
    private void selectNode() {
        result.set(nodesComboBox.getValue());
        hideDialog();
    }

    public SparqlService.NodeInfo getResult() {
        return result.getValue();
    }

    public ObjectProperty<SparqlService.NodeInfo> resultProperty() {
        return result;
    }

    public void showDialog() {
        dialogStage.show();
    }

    public void hideDialog() {
        dialogStage.hide();
    }

    public Stage getDialogStage() {
        return dialogStage;
    }
}
