package be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.webmap;


import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.application.Platform;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.concurrent.Worker.State;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.Node;
import javafx.scene.control.*;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Priority;
import javafx.scene.layout.Region;
import javafx.scene.web.WebEngine;
import javafx.scene.web.WebView;
import javafx.util.Duration;
import netscape.javascript.JSObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Jasper Potts
 */
public class DrawingWebMap extends BorderPane {
    private static final Logger LOG = LoggerFactory.getLogger(DrawingWebMap.class);

    private Timeline locationUpdateTimeline;

    private boolean updatingRectangle = false;
    private final WebView webView;
    final WebEngine webEngine;

    private boolean initialized = false;

    public static class MapRectangle {
        private final double minLat, maxLat, minLon, maxLon;

        public MapRectangle(double minLat, double maxLat, double minLon, double maxLon) {
            this.minLat = minLat;
            this.maxLat = maxLat;
            this.minLon = minLon;
            this.maxLon = maxLon;
        }

        public double getMinLat() {
            return minLat;
        }

        public double getMaxLat() {
            return maxLat;
        }

        public double getMinLon() {
            return minLon;
        }

        public double getMaxLon() {
            return maxLon;
        }
    }

    private final ObjectProperty<MapRectangle> rectangle = new SimpleObjectProperty<>(null);

    public DrawingWebMap() {
        // create web engine and view
        webView = new WebView();
        webEngine = webView.getEngine();
        webEngine.getLoadWorker().stateProperty().addListener(
                (ov, oldState, newState) -> {
                    if (newState == State.SUCCEEDED) {
                        JSObject win = (JSObject) webEngine.executeScript("window");
                        win.setMember("app", new JavaToMapBindingApp());

                        webEngine.executeScript("initialize()");

                        initialized = true;

                        if(rectangle.get() != null)
                            updateRectangleOnMap();
                    }
                }
        );

        webEngine.load(getClass().getResource("drawinggooglemap.html").toString());

        // create map type buttons
        final ToggleGroup mapTypeGroup = new ToggleGroup();
        final ToggleButton road = new ToggleButton("Road");
        road.setSelected(true);
        road.setToggleGroup(mapTypeGroup);
        final ToggleButton satellite = new ToggleButton("Satellite");
        satellite.setToggleGroup(mapTypeGroup);
        final ToggleButton hybrid = new ToggleButton("Hybrid");
        hybrid.setToggleGroup(mapTypeGroup);
        final ToggleButton terrain = new ToggleButton("Terrain");
        terrain.setToggleGroup(mapTypeGroup);
        mapTypeGroup.selectedToggleProperty().addListener((observableValue, toggle, toggle1) -> {
            if (road.isSelected()) {
                webEngine.executeScript("document.setMapTypeRoad()");
            } else if (satellite.isSelected()) {
                webEngine.executeScript("document.setMapTypeSatellite()");
            } else if (hybrid.isSelected()) {
                webEngine.executeScript("document.setMapTypeHybrid()");
            } else if (terrain.isSelected()) {
                webEngine.executeScript("document.setMapTypeTerrain()");
            }
        });

        webEngine.setOnAlert(text ->
                new Alert(Alert.AlertType.WARNING, text.toString()).show());


        // add search
        final TextField searchBox = new TextField();
        //searchBox.setColumns(12);
        searchBox.setPromptText("Search");
        searchBox.textProperty().addListener((observableValue, s, s1) -> {
            // delay location updates to we don't go too fast file typing
            if (locationUpdateTimeline != null) {
                locationUpdateTimeline.stop();
            }

            final Duration initialDelay = new Duration(400);
            final EventHandler<ActionEvent> performSearch = new EventHandler<ActionEvent>() {
                @Override
                public void handle(ActionEvent actionEvent) {
                    webEngine.executeScript("document.goToLocation(\"" + searchBox.getText() + "\")");
                }
            };

            locationUpdateTimeline = new Timeline(new KeyFrame(initialDelay, performSearch));
            locationUpdateTimeline.play();
        });

        Button zoomIn = new Button("Zoom In");
        zoomIn.setOnAction(actionEvent -> webEngine.executeScript("document.zoomIn()"));
        Button zoomOut = new Button("Zoom Out");
        zoomOut.setOnAction(actionEvent -> webEngine.executeScript("document.zoomOut()"));

        //workaround voor een bug die soms voorkomt (bug tussen google maps api en javafx 2.2)
        webView.setOnDragDetected(t -> {
            // intentionally do nothing
        });

        // create toolbar
        ToolBar toolBar = new ToolBar();
        toolBar.getStyleClass().add("map-toolbar");
        toolBar.getItems().addAll(
                road, satellite, hybrid, terrain,
                //                createSpacer(),
                //                google, yahoo, bing,
                createSpacer(),
                new Label("Location:"), searchBox, zoomIn, zoomOut);
        // create root
        getStyleClass().add("map");
        getStylesheets().add(getClass().getResource("WebMap.css").toString());
        setCenter(webView);
        setTop(toolBar);
    }

    private double roundCoordinate(double coord) {
        return Math.round(coord * 10_000) / 10_000.0;
    }

    @SuppressWarnings("unused") //these functions are called from Javascript!
    public class JavaToMapBindingApp {

        public void rectangleValueUpdated(double minLat, double maxLat, double minLon, double maxLon) {
            if (updatingRectangle)
                return;
            rectangle.setValue(new MapRectangle(roundCoordinate(minLat), roundCoordinate(maxLat),
                    roundCoordinate(minLon), roundCoordinate(maxLon)));
        }

        public void clearRectangle() {
            rectangle.setValue(null);
        }

        public void log(String message) {
            LOG.info(message);
        }
    }

    private Node createSpacer() {
        Region spacer = new Region();
        HBox.setHgrow(spacer, Priority.ALWAYS);
        return spacer;
    }

    public void addMarker(final double lat, final double lon) {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                webEngine.executeScript("document.addMarker(" + lat + ", " + lon + ")");
            }

        });
    }

    public void addMarkers() {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                webEngine.executeScript("document.addMarkers()");
            }

        });
    }

    public void addRectangle(double lat1, double lat2, double lon1, double lon2) {
        webEngine.executeScript("document.addRectangle(" + lat1 + ", " + lat2 + ", " + lon1 + ", " + lon2 + ")");
    }

    public void deleteMarkers() {
        webEngine.executeScript("document.deleteOverlays()");
    }

    static { // use system proxy settings when standalone application    
        System.setProperty("java.net.useSystemProxies", "true");
    }

    public void clearRectangle() {
        webEngine.executeScript("document.clearRectangle()");
    }

    public void clearRectangleDrawing() {
        webEngine.executeScript("document.clearRectangleDrawing()");
    }

    public MapRectangle getRectangle() {
        return rectangle.get();
    }

    public ReadOnlyObjectProperty<MapRectangle> rectangleProperty() {
        return rectangle;
    }

    public void setRectangle(MapRectangle rectangle) {
        this.rectangle.set(rectangle);

        updateRectangleOnMap();
    }

    private void updateRectangleOnMap() {
        //only update when webmap is ready
        if (initialized) {
            updatingRectangle = true;

            MapRectangle rectangle = this.rectangle.get();
            assert (rectangle != null);
            webEngine.executeScript("document.setRectangle(" + rectangle.getMinLat() + ", " + rectangle.getMaxLat() + ", " + rectangle.getMinLon() + ", " + rectangle.getMaxLon() + ")");

            updatingRectangle = false;
        }
    }
}
