package be.iminds.ilabt.jfed.experimenter_gui.ext_requests;

import be.iminds.ilabt.jfed.ApplicationParametersModule;
import be.iminds.ilabt.jfed.experimenter_gui.login.URLLogin;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.InvalidLoginException;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.PlanetlabUserLoginModel;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.UserLoginModelManager;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.IOUtils;
import javafx.stage.Stage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Provider;
import javax.xml.bind.DatatypeConverter;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.*;
import java.util.Map;

/**
 * User: twalcari
 * Date: 6/26/2015
 * Time: 11:07
 */
public class FirstLoginParametersHandler implements ParameterHandler<FirstLoginParametersHandler.LoginResult> {
    private static final Logger LOG = LoggerFactory.getLogger(FirstLoginParametersHandler.class);
    private static final String SCHEME_FILE = "file";

    private final UserLoginModelManager userLoginModelManager;
    private final Provider<Stage> parentStageProvider;
    private final ApplicationParametersModule.ApplicationParameters applicationParameters;

    @Inject
    public FirstLoginParametersHandler(UserLoginModelManager userLoginModelManager,
                                       @Named("parentStage") Provider<Stage> parentStageProvider,
                                       ApplicationParametersModule.ApplicationParameters applicationParameters) {
        this.userLoginModelManager = userLoginModelManager;
        this.parentStageProvider = parentStageProvider;
        this.applicationParameters = applicationParameters;
    }


    public static class LoginResult {
        private boolean isPlanetLab = false;
        private boolean loginDefined = false;
        private boolean loggedIn = false;

        public boolean isPlanetLab() {
            return isPlanetLab;
        }

        public boolean isLoginDefined() {
            return loginDefined;
        }

        public boolean isLoggedIn() {
            return loggedIn;
        }

        public void setLoggedIn(boolean loggedIn) {
            this.loggedIn = loggedIn;
        }

        public void setIsPlanetLab(boolean isPlanetLab) {
            this.isPlanetLab = isPlanetLab;
        }

        public void setLoginDefined(boolean loginDefined) {
            this.loginDefined = loginDefined;
        }
    }


    @Override
    public LoginResult handleParameters() {
        Map<String, String> params = applicationParameters.getParameters();
        LOG.debug("Handling FirstLoginParameters");
        LoginResult result = new LoginResult();

        //check if we got login-information via the parameters
        if (params.containsKey(StartupParameters.PARAM_LOGIN_CERTIFICATE)) {
            String loginCertificateLocation = params.get(StartupParameters.PARAM_LOGIN_CERTIFICATE);

            LOG.info("Found login-certificate in parameters. URI={}", loginCertificateLocation);
            try {
                URI uri = new URI(loginCertificateLocation);
                tryLoginWithUri(uri, result);
            } catch (URISyntaxException ex) {
                LOG.error("Could not parse provided login certificate location", ex);
            }
        }

        if (params.containsKey(StartupParameters.PARAM_LOGIN_CERTIFICATE_STRING)) {
            if (!result.isLoginDefined()) {
                String loginCertificateString =
                        params.get(StartupParameters.PARAM_LOGIN_CERTIFICATE_STRING);

                LOG.info("Found login-certificate string in parameters. size={}", loginCertificateString.length());

                try {
                    byte[] decodedLoginCertificate = DatatypeConverter.parseBase64Binary(loginCertificateString);

                    String decodedLoginCertificateString = new String(decodedLoginCertificate);
                    LOG.trace("Base64-decoded login certificate (size={})", decodedLoginCertificateString.length());

                    tryLoginWithString(decodedLoginCertificateString, result);
                } catch (IllegalArgumentException ex) {
                    LOG.warn("Could not Base64-decode login-certificate-string", ex);
                }
            } else {
                LOG.warn("Already logged in! Not processing login certificate string-parameter!");
            }
        }

        if (params.containsKey(StartupParameters.PARAM_LOGIN_HASH)) {

            if (!result.isLoginDefined()) {
                File loginCertFile = URLLogin.getLoginCertificateFile(params.get(StartupParameters.PARAM_LOGIN_HASH));

                if (loginCertFile.exists()) {
                    LOG.info("Found cached certificate with hash {}", params.get(StartupParameters.PARAM_LOGIN_HASH));
                    tryLoginWithUri(loginCertFile.toURI(), result);
                } else {
                    LOG.info("Could not find cached login certificate with hash {}", params.get(StartupParameters.PARAM_LOGIN_HASH));
                }
            } else {

                LOG.warn("Already logged in! Not processing login hash-parameter!");
            }

        }


        //now check if we can login without a password, or with the provided password

        if (result.isLoginDefined()) {
            if (!result.isPlanetLab()) {
                if (!userLoginModelManager.getKeyCertUserLoginModel().isPasswordRequired()) {
                    userLoginModelManager.login();

                    result.setLoggedIn(true);
                } else if (params.containsKey(StartupParameters.PARAM_LOGIN_PASSWORD)) {
                    if (userLoginModelManager.getKeyCertUserLoginModel().unlock(params.get(StartupParameters.PARAM_LOGIN_PASSWORD))) {
                        userLoginModelManager.login();

                        result.setLoggedIn(true);
                    } else {
                        LOG.debug("Could not login with provided password");
                    }
                }
            } else {
                if (userLoginModelManager.getPlanetlabUserLoginModel().isReadyToFetchCertificate()) {
                    //first try to fetch the login certificate before continuing
                    try {
                        userLoginModelManager.getPlanetlabUserLoginModel().fetchCertificate();

                        if (!userLoginModelManager.getPlanetlabUserLoginModel().isPasswordRequired()) {
                            userLoginModelManager.login();

                            result.setLoggedIn(true);
                        } else if (params.containsKey(StartupParameters.PARAM_LOGIN_PASSWORD)) {
                            if (userLoginModelManager.getPlanetlabUserLoginModel().unlock(params.get(StartupParameters.PARAM_LOGIN_PASSWORD))) {
                                userLoginModelManager.login();
                                result.setLoggedIn(true);
                            } else {
                                LOG.debug("Could not login with provided password");
                            }
                        }
                    } catch (PlanetlabUserLoginModel.CertificateFetchException ex) {
                        JFDialogs.create()
                                .owner(parentStageProvider.get())
                                .message("jFed could not fetch the user certificate from the authority. Contact support for more information.")
                                .masthead("Login Failed")
                                .title("Login Failed")
                                .showException(ex);
                    }
                }
            }


        }

        return result;
    }

    private boolean tryLoginWithUri(URI uri, LoginResult result) {
        //try to update KeyCertUserLoginModel with the new information
        if (uri.getScheme().equals(SCHEME_FILE)) {
            File loginCertificateFile = new File(uri);

            //try to detect type of certificate
            try {
                String fileContents = IOUtils.fileToString(loginCertificateFile);
                result.setIsPlanetLab(fileContents.startsWith("fed4fire"));
            } catch (IOException ex) {
                LOG.warn("Could not detect the type of login file: error reading provided file", ex);
            }

            if (!result.isPlanetLab()) {
                try {
                    userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO);
                    userLoginModelManager.getKeyCertUserLoginModel().setKeyCertPemFile(loginCertificateFile);
                    userLoginModelManager.save();

                    result.setLoginDefined(true);
                    return true;
                } catch (InvalidLoginException e) {
                    LOG.error("Could not properly load the provided key-certificate-pair into jFed", e);
                    return false;
                }
            } else {
                try {
                    userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.PLANETLAB);
                    userLoginModelManager.getPlanetlabUserLoginModel().setSshPrivateKeyFile(loginCertificateFile);
                    userLoginModelManager.save();

                    result.setLoginDefined(true);
                    return true;
                } catch (InvalidLoginException e) {
                    LOG.error("Could not properly load the provided planetlab login file into jFed", e);
                    return false;
                }
            }

        } else {
            //try to fetch from the internet
            URL url;
            try {
                url = uri.toURL();
            } catch (MalformedURLException e) {
                LOG.error("Got an illegal URL, cannot continue", e);
                return false;
            }

            //try to detect the type of certificate
            try {
                URLConnection conn = url.openConnection();
                BufferedReader br = new BufferedReader(new InputStreamReader(conn.getInputStream()));

                String firstLine = br.readLine();
                br.close();

                result.setIsPlanetLab(firstLine.startsWith("fed4fire"));

            } catch (IOException ex) {
                LOG.warn("Could not detect the type of login file: error reading provided url", ex);
            }

            if (!result.isPlanetLab()) {
                try {
                    userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO);
                    userLoginModelManager.getKeyCertUserLoginModel().setKeyCertPemUrl(url);
                    userLoginModelManager.save();

                    result.setLoginDefined(true);
                    return true;
                } catch (InvalidLoginException e) {
                    LOG.error("Could not properly load the provided key-certificate-pair into jFed", e);
                    return false;
                }
            } else {
                try {
                    userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.PLANETLAB);
                    userLoginModelManager.getPlanetlabUserLoginModel().setSshPrivateKeyUrl(url);
                    userLoginModelManager.save();

                    result.setLoginDefined(true);
                    return true;
                } catch (InvalidLoginException e) {
                    LOG.error("Could not properly load the provided planetlab-string into jFed", e);
                    return false;
                }
            }
        }
    }

    private boolean tryLoginWithString(String loginString, LoginResult result) {
        result.setIsPlanetLab(loginString.startsWith("fed4fire"));

        if (result.isPlanetLab()) {
            //this is a planetlab file!
            userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.PLANETLAB);

            try {
                userLoginModelManager.getPlanetlabUserLoginModel().setSshPrivateKeyString(loginString);

                result.setLoginDefined(true);
                return true;
            } catch (InvalidLoginException e) {
                LOG.error("Could not properly load the provided planetlab-string into jFed", e);
                return false;
            }

        } else {
            //assume this is a key-cert-combo

            userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO);
            try {
                userLoginModelManager.getKeyCertUserLoginModel().setKeyCertPemString(loginString);

                result.setLoginDefined(true);
                return true;

            } catch (InvalidLoginException e) {
                LOG.error("Could not properly load the provided key-certificate-pair-string into jFed", e);
                userLoginModelManager.load();
                return false;
            }
        }
    }
}
