package be.iminds.ilabt.jfed.experimenter_gui.login;

import be.iminds.ilabt.jfed.connectivity_tester.ConnectivityTest;
import be.iminds.ilabt.jfed.connectivity_tester.SslHostAndPortTest;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.ThreadingAwareDialogsUtil;
import be.iminds.ilabt.jfed.util.ProxyPreferencesManager;
import javafx.scene.control.ButtonType;
import javafx.stage.Stage;
import org.slf4j.LoggerFactory;

import java.net.URL;
import java.util.Date;
import java.util.Optional;

/**
 * LoginConnectivityTest
 */
public abstract class LoginConnectivityTest {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(LoginConnectivityTest.class);

    private final SfaAuthority userAuthority;
    private final ProxyPreferencesManager proxyPrefMan;
    private final GeniUserProvider geniUserProvider;
    private final AuthorityList authorityList;
    private final JFedGuiPreferences jFedPreferences;



    public LoginConnectivityTest(SfaAuthority userAuthority, ProxyPreferencesManager proxyPrefMan,
                                 GeniUserProvider geniUserProvider, AuthorityList authorityList,
                                 JFedGuiPreferences jFedPreferences) {
        this.userAuthority = userAuthority;
        this.proxyPrefMan = proxyPrefMan;
        this.geniUserProvider = geniUserProvider;
        this.authorityList = authorityList;
        this.jFedPreferences = jFedPreferences;
    }

    //used to show feedback messages. implementers should note these can be called on any thread.
    protected abstract void busyMessage(String message);

    protected abstract void okMessage(String message);

    protected abstract void errorMessage(String message);

    /**
     * called when all testing is finished.  (so to use this testing, create a class extending this, and implement this method)
     *
     * @param success    true when connectivity is ok, false otherwise.
     * @param abortLogin true when the user wants to abort login (and login with another account). false when the user wants to login, even though the connectivity test failed.
     */
    protected abstract void connectivityTestFinished(boolean success, boolean abortLogin);

    /**
     * get a javafx Stage for showing dialogs.
     */
    protected abstract Stage getDialogStage();

    public void start() {
        Thread cttThread = new Thread(() -> {
            try {
                final boolean proxyAvailable = !userAuthority.getProxies().isEmpty();
                final boolean proxyActivated = proxyPrefMan.isProxyEnabledForJfed();

                if (!proxyAvailable && proxyActivated) {
                    JFDialogs d = JFDialogs.create()
                            .owner(getDialogStage())
                            .message("jFed does not know a proxy for your user authority (" + userAuthority.getName() + ")\n" +
                                    "However, the proxy is activated in the jFed settings\n\n" +
                                    "Do you want to deactivate the proxy in the settings and continue?")
                            .title("Deactivate Proxy?");

                    Optional<ButtonType> response = ThreadingAwareDialogsUtil.showConfirm(d);

                    if (response.filter(r -> r == ButtonType.YES).isPresent()) {
                        LOG.info("Deactivating proxy");
                        jFedPreferences.setString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_JFED,
                                ProxyPreferencesManager.NEVER_USE_PROXY);
                        jFedPreferences.setString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH,
                                ProxyPreferencesManager.NEVER_USE_PROXY);
                        proxyPrefMan.updateProxySettings();
                    } else {
                        connectivityTestFinished(false/*connectivity success*/, true/*abort login*/);
                        return;
                    }
                }

                boolean proxyOk;
                boolean userAmWithoutProxyOk;
                boolean userAmWithProxyOk;

                if (proxyActivated) {
                    busyMessage("Proxy activated. Checking proxy ports...");

                    assert proxyAvailable;
                    proxyOk = blockingTestProxy();

                    if (proxyOk) {
//                            busyMessage("Proxy activated. Reachable proxy ports. Testing connectivity using proxy...");
//                            userAmWithProxyOk = blockingTestConnectivity(true/*useProxy*/);
//                            if (!userAmWithProxyOk) {
//                                busyMessage("Proxy activated. Reachable proxy ports. No connectivity using proxy. Checking connectivity without proxy...");
//                                userAmWithoutProxyOk = blockingTestConnectivity(false/*useProxy*/);
//                            } else
//                                userAmWithoutProxyOk =
//                                        false; //untested, but doesn't matter, proxy works so we'll use it.

                        //don't test any more: if ports are reachable, that's enough
                        userAmWithProxyOk = true;
                        userAmWithoutProxyOk = false; //untested, but doesn't matter, proxy works so we'll use it.
                    } else {
                        busyMessage("Proxy activated. No reachable proxy ports... checking connectivity without proxy...");
                        userAmWithoutProxyOk = blockingTestConnectivity(false/*useProxy*/);
                        userAmWithProxyOk = false;
                    }

                    if (proxyOk && userAmWithProxyOk) {
                        okMessage("Successfully connected using proxy.");
                        connectivityTestFinished(true/*connectivity success*/, false/*abort login*/);
                        return;
                    }

                    if (!userAmWithProxyOk && userAmWithoutProxyOk) {
                        JFDialogs d = JFDialogs.create()
                                .owner(getDialogStage())
                                .message("jFed could not contact the server using the proxy, but was able to contact it without proxy.\n" +
                                        "Do you want to deactivate the proxy and continue?")
                                .title("Deactivate Proxy?");

                        Optional<ButtonType> response = ThreadingAwareDialogsUtil.showConfirm(d);
                        boolean deactivateAndContinue = response.filter(r -> r == ButtonType.YES).isPresent();
                        if (deactivateAndContinue) {
                            LOG.info("Deactivating proxy");
                            jFedPreferences.setString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_JFED,
                                    ProxyPreferencesManager.NEVER_USE_PROXY);
                            jFedPreferences.setString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH,
                                    ProxyPreferencesManager.NEVER_USE_PROXY);
                            proxyPrefMan.updateProxySettings();
                        }
                        connectivityTestFinished(true/*connectivity success*/, !deactivateAndContinue/*abort login*/);
                        return;
                    }
                } else {
                    if (proxyAvailable)
                        busyMessage("Proxy not activated. Testing connectivity...");
                    else
                        busyMessage("Testing connectivity...");
                    userAmWithoutProxyOk = blockingTestConnectivity(false/*useProxy*/);

                    if (!userAmWithoutProxyOk) {
                        if (proxyAvailable) {
                            busyMessage("No connectivity. Checking proxy ports...");
                            proxyOk = blockingTestProxy();
                            if (proxyOk) {
                                //Do test proxy further in this case.
                                busyMessage("No connectivity. Reachable proxy ports. Testing connectivity using proxy...");
                                userAmWithProxyOk = blockingTestConnectivity(true/*useProxy*/);
                            } else {
                                userAmWithProxyOk = false;
                            }
                        } else {
                            //busyMessage("Proxy not activated. No connectivity.");
                            proxyOk = false;
                            userAmWithProxyOk = false;
                        }
                    } else {
                        //not tested, bu we don't care as it works without
                        proxyOk = false;
                        userAmWithProxyOk = false;
                    }

                    if (userAmWithoutProxyOk) {
                        okMessage("Successfully connected.");
                        connectivityTestFinished(true/*connectivity success*/, false/*abort login*/);
                        return;
                    }

                    if (!userAmWithoutProxyOk && proxyOk && userAmWithProxyOk) {

                        JFDialogs d = JFDialogs.create()
                                .owner(getDialogStage())
                                .message("jFed could not contact the server without the proxy, but was able to contact it with proxy.\n\n" +
                                        "Do you want to activate the proxy and continue?")
                                .title("Activate Proxy?");

                        Optional<ButtonType> response = ThreadingAwareDialogsUtil.showConfirm(d);
                        boolean activateAndContinue = response.filter(r -> r== ButtonType.YES).isPresent();
                        if (activateAndContinue) {
                            LOG.info("Activating proxy");
                            jFedPreferences.setString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_JFED,
                                    ProxyPreferencesManager.ALWAYS_USE_PROXY);
                            jFedPreferences.setString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH,
                                    ProxyPreferencesManager.ALWAYS_USE_PROXY);
                            proxyPrefMan.updateProxySettings();
                        }
                        connectivityTestFinished(true/*connectivity success*/, !activateAndContinue/*abort login*/);
                        return;
                    }
                }

                errorMessage("Connectivity Problem");
                Optional<ButtonType> dialogResponse;
                if (proxyAvailable) {
                    JFDialogs d = JFDialogs.create()
                            .owner(getDialogStage())
                            .message("jFed could not contact the server with or without using the proxy.\n" +
                                    (proxyActivated ?
                                            "Note: The jFed proxy is activated in your settings and will stay activated.\n\n"
                                            :
                                                    "Note: The jFed proxy is deactivated in your settings and will stay deactivated.\n\n") +
                                    "Do you want to continue anyway? (Most likely, nothing will work)")
                            .title("Connectivity Problem");

                    dialogResponse = ThreadingAwareDialogsUtil.showConfirm(d);
                } else {

                    JFDialogs d = JFDialogs.create()
                            .owner(getDialogStage())
                            .message("jFed could not contact the server.\n" +
                                    "Note: jFed knows no proxy for your user authority (" + userAuthority.getName() + "), so no proxy was tried.\n\n" +
                                    "Do you want to continue anyway? (Most likely, nothing will work)")
                            .title("Connectivity Problem");
                    dialogResponse = ThreadingAwareDialogsUtil.showConfirm(d);

                }
                boolean continueLogin = dialogResponse.filter(r -> r == ButtonType.YES).isPresent();
                connectivityTestFinished(false/*connectivity success*/, !continueLogin/*abort login*/);
            } catch (Exception | AssertionError e) {
                LOG.error("Exception while testing connectivity", e);
                errorMessage("Exception while testing connectivity: " + e.getMessage());
                connectivityTestFinished(false/*connectivity success*/, true/*abort login*/);
            }
        });
        cttThread.setName("ConnectivityTestThread");
        cttThread.setDaemon(true);
        cttThread.start();
    }

    private boolean proxyEnabledInPreferences() {
        return proxyPrefMan.isProxyEnabledForJfed();
    }

    private boolean blockingTestProxy() {
        busyMessage("Testing connectivity to proxy...");

        final ProxyPreferencesManager.TryProxyTask tryProxyTask = proxyPrefMan.getTryProxyTask(userAuthority);
        final Thread thread = new Thread(tryProxyTask);
        thread.setName("ProxyTestThread");
        thread.setDaemon(true);
        thread.start();

        try {
            thread.join(60 * 1000); //wait maximum 1 minute for this test.
        } catch (InterruptedException e) {
            LOG.error("Thread " + thread.getName() + " was interrupted.", e);
        }

        if (thread.isAlive()) {
            errorMessage("Proxy Test Task takes too long. Assuming failure.");
            LOG.error("Proxy Test Task takes too long. Assuming failure. (" + tryProxyTask.getState() + "). " + tryProxyTask.getMessage(), tryProxyTask.getException());
            return false;
        }
        if (!tryProxyTask.isFinished()) {
            errorMessage("Proxy Test Task Failed.");
            LOG.error("Proxy Test Task Failed (" + tryProxyTask.getState() + "). " + tryProxyTask.getMessage(), tryProxyTask.getException());
            return false;
        }

        boolean res = tryProxyTask.isSuccess();
        proxyPrefMan.updateProxySettings();
        return res;
    }

    private final class Helper<T> {
        public T t;

        public Helper(T t) {
            this.t = t;
        }
    }

    private boolean blockingTestConnectivity(boolean useProxy) {
        if (userAuthority.getUrls().isEmpty()) {
            LOG.warn("Could not find URL for AM to test connectivity");
            return false;
        }

        //choose a URL to contact. Prefer AM's
        URL url = userAuthority.getUrls().get(new ServerType(ServerType.GeniServerRole.AM, 2));
        if (url == null) url = userAuthority.getUrls().get(new ServerType(ServerType.GeniServerRole.AM, 3));
        if (url == null) url = userAuthority.getUrls().get(new ServerType(ServerType.GeniServerRole.AM, 4));
        if (url == null) url = userAuthority.getUrls().get(new ServerType(ServerType.GeniServerRole.AM, 1));
        if (url == null) url = userAuthority.getUrls().values().iterator().next();
        LOG.debug("Trying to contact '{}' for connectivity test.", url.toExternalForm());

        //sslHostAndPortTest.call() seems to hang on MAC. So we'll have to manually wait and interrupt it if it takes to long
        JFedConnection.ProxyInfo proxyInfo;
        if (useProxy) {
            //commented code below doesn't work: it relies on proxy being activated in settings
//            //get the same proxy that will be used for all connections
//            proxyPrefMan.updateProxySettings();
//            proxyInfo = experimenterModel.getSfaConnectionPool().getProxyInfo();

            GeniUser currentUser = geniUserProvider.getLoggedInGeniUser();
            proxyInfo =
                    proxyPrefMan.chooseUserProxy(userAuthority.getProxies(), currentUser, new UserSshKeyInfo(currentUser));

            if (proxyInfo == null)
                LOG.warn("useProxy=true but proxyInfo=null");
        } else
            proxyInfo = null;

        final SslHostAndPortTest sslHostAndPortTest = new SslHostAndPortTest(url, null, null, proxyInfo, authorityList);
        final Helper<ConnectivityTest.ConnectivityTestResult> connectivityTestResultHelper =
                new Helper<>(null);

        Runnable run = () -> {
            try {
                connectivityTestResultHelper.t = sslHostAndPortTest.call();
            } catch (Exception e) {
                LOG.warn("Exception in sslHostAndPortTest.call(): " + e.getMessage(), e);
                connectivityTestResultHelper.t = null;
            }
        };
        Thread conTestThread = new Thread(run);
        Date startTime = new Date();
        Date stopTime;
        conTestThread.start();
        try {
            conTestThread.join(22 * 1000); //wait 2 seconds longer than connection timeouts  (2 times 10s)
            stopTime = new Date();
        } catch (InterruptedException e) {
            LOG.warn("conTestThread.join(12*1000); was interrupted", e);
            stopTime = new Date();
        }
        long duration_ms = stopTime.getTime() - startTime.getTime();
        LOG.info("sslHostAndPortTest.call() took " + duration_ms + "ms. alive=" + conTestThread.isAlive() + " result=" + connectivityTestResultHelper.t);
        //if timeout occurred, stop thread (or at least try to)
        if (conTestThread.isAlive()) {
            conTestThread.interrupt();
        }

        final ConnectivityTest.ConnectivityTestResult connectivityTestResult = connectivityTestResultHelper.t;

        if (connectivityTestResult == null || connectivityTestResult.getStatus() != ConnectivityTest.Status.SUCCEEDED) {
            LOG.warn("Could not establish SSL-connection to {} with proxyInfo {}", url, proxyInfo);
            return false;
        }

        LOG.info("Successfully connected to the AM in connectivity test. useProxy=" + useProxy);

        return true;
    }
}
