package be.iminds.ilabt.jfed.lowlevel.api;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.*;

/**
 * @author Wim Van de Meerssche
 * @see <a href="http://groups.geni.net/geni/wiki/GAPI_AM_API_V2">Geni Aggregate Manager API v2</a>
 */
public class ProtoGeniAMExtensions extends AbstractGeniAggregateManager {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ProtoGeniAMExtensions.class);

    /**
     * Construct a new ProtoGeniAMExtensions, with a certain logger.
     * <p/>
     * <p>This class stores no internal state, except for the logger used. It sends all results and debug info to this {@link be.iminds.ilabt.jfed.log.Logger} if it is non null.
     *
     * @param logger        the logger to use. May be null if no logger is used.
     * @param retrySettings the retrySettings to use, or null to use the static globalRetrySettings in AbstractApi
     */
    public ProtoGeniAMExtensions(@Nonnull Logger logger, RetrySettings retrySettings, @Nonnull JFedPreferences jFedPreferences) {
        super(logger, retrySettings, new ServerType(ServerType.GeniServerRole.AM, 2), jFedPreferences);
    }

    /**
     * Construct a new ProtoGeniAMExtensions, with a certain logger.
     * <p/>
     * <p>This class stores no internal state, except for the logger used. It sends all results and debug info to this {@link be.iminds.ilabt.jfed.log.Logger} if it is non null.
     * <p/>
     * <p>This constructor sets autoRetryBusy to true, so "busy" replies are retried.
     *
     * @param logger the logger to use. May be null if no logger is used.
     */
    public ProtoGeniAMExtensions(@Nonnull Logger logger, @Nonnull JFedPreferences jFedPreferences) {
        super(logger, null, new ServerType(ServerType.GeniServerRole.AM, 2), jFedPreferences);
    }

    /**
     * A human readable name for the implemented API
     *
     * @return "ProtoGeni Aggregate Manager API Extensions";
     */
    static public String getApiName() {
        return "ProtoGeni Aggregate Manager API Extensions";
    }

    private static void addHashtables(Hashtable target, Map<String, Object> source) {
        if (source == null)
            return;
        for (Object o : source.keySet())
            target.put(o, source.get(o));
    }

    /**
     * A human readable name for the implemented API
     *
     * @return "ProtoGeni Aggregate Manager API Extensions";
     */
    @Override
    public String getName() {
        return getApiName();
    }

    @Override
    public AggregateManagerReply<? extends AbstractVersionInfo> getVersion(SfaConnection con) throws JFedException {
        throw new RuntimeException("not implemented here");
    }

    /**
     * ListImages
     * <p/>
     * example return
     * "value": [
     * {
     * "urn": "urn:publicid:IDN+wall2.ilabt.iminds.be+image+wall2-ilabt-iminds-be:testImage",
     * "url": "https://www.wall2.ilabt.iminds.be/image_metadata.php?uuid=adf41cd4-c3dd-11e3-b407-001517becdc1"
     * }
     * ]
     */
    @ApiMethod(order = 1, hint = "ListImages call")
    public AggregateManagerReply<List<ImageInfo>> listImages(SfaConnection con,
                                                             @ApiMethodParameter(name = "credentialList", hint = "mandatory list of credentials. this must include a valid user credential.")
                                                             List<AnyCredential> credentialList,
                                                             @ApiMethodParameter(name = "userUrn", required = true, hint = "user to list images for", parameterType = ApiMethodParameterType.USER_URN)
                                                             String userUrn,
                                                             @ApiMethodParameter(name = "extraOptions", hint = "extra options", required = false, guiDefaultOptional = false, parameterType = ApiMethodParameterType.GENI_EXTRA_OPTIONS)
                                                             Map<String, Object> extraOptions) throws JFedException {
        Map<String, Object> methodParams =
                makeMethodParameters("credentialList", credentialList, "userUrn", userUrn, "extraOptions", extraOptions);

        checkNonNullArgument(credentialList, "credentialList");
        checkNonNullArgument(userUrn, "userUrn");

        Vector credentialsVector = new Vector();
        credentialsVector.setSize(credentialList.size());
        for (int i = 0; i < credentialList.size(); i++) {
            AnyCredential c = credentialList.get(i);
            if (c == null)
                throw new IllegalArgumentException("Illegal arguments: no AnyCredential is allowed to be null in credentialList: " + credentialList);
            String credXml = c.getCredentialXml();
            if (credXml == null)
                throw new IllegalArgumentException("Illegal arguments: no AnyCredential is allowed to have null as XML in credentialList: " + credentialList);
            credentialsVector.set(i, credXml);
        }

        Hashtable options = new Hashtable();
        addHashtables(options, extraOptions);

        Vector argv = new Vector();
        argv.setSize(3);
        argv.set(0, userUrn);
        argv.set(1, credentialsVector);
        argv.set(2, options);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "ListImages", argv, methodParams);

        List<ImageInfo> listImageResult = null;
        if (AggregateManagerReply.isSuccess(res))
            try {
                listImageResult = new ArrayList<ImageInfo>();
                Vector<Hashtable> v = apiSpecifiesVectorOfT(Hashtable.class, res.getResultValueObject());
                for (Hashtable ht : v) {
                    Hashtable<String, String> htss = apiSpecifiesHashtableStringToString(ht);
                    ImageInfo imageInfo = new ImageInfo(htss);
                    listImageResult.add(imageInfo);
                }
            } catch (Exception e) {
                LOG.warn("Exception parsing ListImage result: " + e.getMessage(), e);
                listImageResult = null;
            }

        AggregateManagerReply<List<ImageInfo>> r = new AggregateManagerReply<List<ImageInfo>>(res, listImageResult);
        log(res, r, "listImages", "ListImages", con, methodParams);
        return r;
    }

    /**
     * DeleteImage
     * <p/>
     * example return:
     * "     value": true
     */
    @ApiMethod(order = 2, hint = "DeleteImage call")
    public AggregateManagerReply<Boolean> deleteImage(SfaConnection con,
                                                      @ApiMethodParameter(name = "credentialList", hint = "mandatory list of credentials. this must include a valid user credential.")
                                                      List<AnyCredential> credentialList,
                                                      @ApiMethodParameter(name = "imageUrn", required = true, hint = "image URN to delete", parameterType = ApiMethodParameterType.URN)
                                                      String imageUrn,
                                                      @ApiMethodParameter(name = "creatorUrn", required = false, hint = "Optional: creator user URN, needed if image not owned by calling user", parameterType = ApiMethodParameterType.USER_URN)
                                                      String creatorUrn,
                                                      @ApiMethodParameter(name = "extraOptions", hint = "extra options", required = false, guiDefaultOptional = false, parameterType = ApiMethodParameterType.GENI_EXTRA_OPTIONS)
                                                      Map<String, Object> extraOptions) throws JFedException {
        Map<String, Object> methodParams =
                makeMethodParameters("credentialList", credentialList, "imageUrn", imageUrn, "creatorUrn", creatorUrn, "extraOptions", extraOptions);

        checkNonNullArgument(credentialList, "credentialList");
        checkNonNullArgument(imageUrn, "imageUrn");

        Vector credentialsVector = new Vector();
        credentialsVector.setSize(credentialList.size());
        for (int i = 0; i < credentialList.size(); i++) {
            AnyCredential c = credentialList.get(i);
            if (c == null)
                throw new IllegalArgumentException("Illegal arguments: no AnyCredential is allowed to be null in credentialList: " + credentialList);
            String credXml = c.getCredentialXml();
            if (credXml == null)
                throw new IllegalArgumentException("Illegal arguments: no AnyCredential is allowed to have null as XML in credentialList: " + credentialList);
            credentialsVector.set(i, credXml);
        }

        Hashtable options = new Hashtable();
        if (creatorUrn != null)
            options.put("creator_urn", creatorUrn);
        addHashtables(options, extraOptions);

        Vector argv = new Vector();
        argv.setSize(3);
        argv.set(0, imageUrn);
        argv.set(1, credentialsVector);
        argv.set(2, options);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "DeleteImage", argv, methodParams);

        AggregateManagerReply<Boolean> r = new AggregateManagerReply<Boolean>(res);
        log(res, r, "deleteImage", "DeleteImage", con, methodParams);
        return r;
    }

    /**
     * CreateImage
     */
    @ApiMethod(order = 3, hint = "CreateImage call")
    public AggregateManagerReply<ImageInfo> createImage(SfaConnection con,
                                                        @ApiMethodParameter(name = "credentialList", hint = "mandatory list of credentials.")
                                                        List<AnyCredential> credentialList,
                                                        @ApiMethodParameter(name = "sliceUrn", required = true, hint = "slice URN", parameterType = ApiMethodParameterType.SLICE_URN)
                                                        String sliceUrn,
                                                        @ApiMethodParameter(name = "imageName", required = true, hint = "image name", parameterType = ApiMethodParameterType.STRING)
                                                        String imageName,
                                                        @ApiMethodParameter(name = "sliverUrn", required = true, hint = "sliver URN (= URN of node)", parameterType = ApiMethodParameterType.URN)
                                                        String sliverUrn,
                                                        @ApiMethodParameter(name = "global", required = false, hint = "optional: global flag (if global is true, the image is public, otherwise, it is private)", parameterType = ApiMethodParameterType.BOOLEAN)
                                                        Boolean global,
                                                        @ApiMethodParameter(name = "updatePrepare", required = false,
                                                                hint = "optional: update_prepare flag (if true, the prepare -M option " +
                                                                "is used for image creation, which will cause system users to be saved in the image, default is false)",
                                                                parameterType = ApiMethodParameterType.BOOLEAN, guiDefault = "false")
                                                        Boolean updatePrepare,
                                                        @ApiMethodParameter(name = "extraOptions", hint = "extra options", required = false, guiDefaultOptional = false, parameterType = ApiMethodParameterType.GENI_EXTRA_OPTIONS)
                                                        Map<String, Object> extraOptions) throws JFedException {
        Map<String, Object> methodParams = makeMethodParameters(
                "credentialList", credentialList,
                "sliceUrn", sliceUrn,
                "imageName", imageName,
                "sliverUrn", sliverUrn,
                "global", global,
                "updatePrepare", updatePrepare,
                "extraOptions", extraOptions);

        checkNonNullArgument(credentialList, "credentialList");
        checkNonNullArgument(sliceUrn, "sliceUrn");
        checkNonNullArgument(imageName, "imageName");
        checkNonNullArgument(sliverUrn, "sliverUrn");

        Vector credentialsVector = new Vector();
        credentialsVector.setSize(credentialList.size());
        for (int i = 0; i < credentialList.size(); i++) {
            AnyCredential c = credentialList.get(i);
            if (c == null)
                throw new IllegalArgumentException("Illegal arguments: no AnyCredential is allowed to be null in credentialList: " + credentialList);
            String credXml = c.getCredentialXml();
            if (credXml == null)
                throw new IllegalArgumentException("Illegal arguments: no AnyCredential is allowed to have null as XML in credentialList: " + credentialList);
            credentialsVector.set(i, credXml);
        }

        Hashtable options = new Hashtable();
        if (global != null)
            options.put("global", global);
        if (updatePrepare != null)
            options.put("update_prepare", updatePrepare);
        addHashtables(options, extraOptions);

        Vector argv = new Vector();
        argv.setSize(5);
        argv.set(0, sliceUrn);
        argv.set(1, imageName);
        argv.set(2, sliverUrn);
        argv.set(3, credentialsVector);
        argv.set(4, options);

        XMLRPCCallDetailsWithCodeValueError res = executeXmlRpcCommandGeni(con, "CreateImage", argv, methodParams);

        ImageInfo createImageResult = null;
        if (AggregateManagerReply.isSuccess(res))
            try {
                Vector<String> v = apiSpecifiesVectorOfString(res.getResultValueObject());
                if (v.size() != 2) {
                    if (v.size() < 2)
                        throw new RuntimeException("returned array does not have exactly 2 items. It has " + v.size());
                    else {
                        LOG.warn("returned array of CreateImage does not have exactly 2 items. It has {}. Will try to continue anyway.", v.size());
                    }
                }
                createImageResult = new ImageInfo(v.get(0), v.get(1));
            } catch (Exception e) {
                LOG.warn("Exception parsing CreateImage result: " + e.getMessage(), e);
            }

        AggregateManagerReply<ImageInfo> r = new AggregateManagerReply<ImageInfo>(res, createImageResult);
        log(res, r, "createImage", "CreateImage", con, methodParams);
        return r;
    }

    public static class ImageInfo implements AggregateManagerWrapper.ImageInfo {
        private final String urn;
        private final String url;

        public ImageInfo(String urn, String url) {
            this.urn = urn;
            this.url = url;
        }

        public ImageInfo(Hashtable<String, String> ht) {
            this.urn = ht.get("urn");
            this.url = ht.get("url");
        }

        public String getUrn() {
            return urn;
        }

        public String getUrl() {
            return url;
        }

        @Override
        public String toString() {
            return "ImageInfo{" +
                    "urn='" + urn + '\'' +
                    ", url='" + url + '\'' +
                    '}';
        }
    }

}
