package be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractGeniAggregateManager;
import be.iminds.ilabt.jfed.lowlevel.api.AggregateManager3;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.ErrorDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.lowlevel.stitching.StitchingDirector;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.RFC3339Util;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.ParseException;
import java.util.*;

/**
 * AMv3Wrapper
 */
public class AMv3Wrapper extends AggregateManagerWrapper {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(AMv3Wrapper.class);

    private AggregateManager3 am3;



    /**
     * returns only the SfaCredentials in the list, discarding the ABAC and other credentials.
     *
     * This has been temporary added to AMv3Wrapper. Make this return origCredentials to get the correct MAv3 behaviour. see FEDIBBTDEV-752
     * */
    @Override
    public List<AnyCredential> filterCredentials(List<AnyCredential> origCredentials) {
        if (origCredentials == null)
            return null;

        boolean removeAbac = amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.REMOVE_ABAC_CRED);
        boolean removeSfa = amAuthority.hasSpecialCase(SfaAuthority.SpecialCase.REMOVE_SFA_CRED);

        if (!removeAbac && !removeSfa) {
            //remove nothing
            return origCredentials;
        }

        List<AnyCredential> res = new ArrayList<AnyCredential>();
        for (AnyCredential cred : origCredentials) {
            if (cred instanceof SfaCredential) {
                if (!removeSfa) {
                    res.add(cred);
                }else {
                    LOG.debug("Removing SfaCredential");
                }
            }
            if (cred instanceof AbacCredential) {
                if ((!removeAbac || cred.isSpeaksFor()))
                    res.add(cred);
                else {
                    LOG.debug("Removing AbacCredential");
                }
            }
        }
        if (res.isEmpty() && !origCredentials.isEmpty()) {
            LOG.warn("After filtering credentials, none are left. Will have to fall back to using them all anyway.");
            return origCredentials;
        }
        return res;
    }


    public AMv3Wrapper(Logger logger, GeniUserProvider geniUserProvider, JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences, SfaAuthority amAuthority) {
        super(logger, geniUserProvider, connectionProvider, jFedPreferences, amAuthority);

        am3 = new AggregateManager3(logger, jFedPreferences);
    }

    public static StatusDetails.SliverStatus getGlobalSliverStatus(AggregateManager3.SliverInfo sliverInfo) {
        String operationalStatus = sliverInfo.getOperationalStatus();

        //allocation states come first, only for the "geni_provisioned" state, we need to look at the operational state
        if (sliverInfo.getAllocationStatus().equalsIgnoreCase(AggregateManager3.GENI_UNALLOCATED_STATE))
            return StatusDetails.SliverStatus.UNALLOCATED;
        if (sliverInfo.getAllocationStatus().equalsIgnoreCase(AggregateManager3.GENI_ALLOCATED_STATE))
            return StatusDetails.SliverStatus.CHANGING; //actually, this means "unallocated" in SliverStatus terminology :-)
        //        if (sliverInfo.getAllocationStatus().equalsIgnoreCase("geni_provisioned")) return SliverStatus.CHANGING;

        //AM API v3 defined operatinoal states  (AMv3Wrapper does not currently support extra operational states advertised in RSpec advertisement)
        if (operationalStatus.equalsIgnoreCase("geni_pending_allocation"))
            return StatusDetails.SliverStatus.CHANGING;
        if (operationalStatus.equalsIgnoreCase("geni_notready")) return StatusDetails.SliverStatus.CHANGING;
        if (operationalStatus.equalsIgnoreCase("geni_configuring")) return StatusDetails.SliverStatus.CHANGING;
        if (operationalStatus.equalsIgnoreCase("geni_stopping")) return StatusDetails.SliverStatus.CHANGING;
        if (operationalStatus.equalsIgnoreCase("geni_ready")) return StatusDetails.SliverStatus.READY;
        if (operationalStatus.equalsIgnoreCase("geni_ready_busy")) return StatusDetails.SliverStatus.READY;
        if (operationalStatus.equalsIgnoreCase("geni_failed")) return StatusDetails.SliverStatus.FAIL;

        LOG.warn("Server returned unknown sliver operationalStatus: \"" + operationalStatus + "\"  (allocationStatus=" + sliverInfo.getAllocationStatus() + ")");
        return StatusDetails.SliverStatus.UNKNOWN;
    }

    protected SfaConnection getConnection() throws JFedException {
        return (SfaConnection) getConnection(new ServerType(ServerType.GeniServerRole.AM, 3));
    }

    @Override
    public Hashtable getVersion() throws JFedException {
        AbstractGeniAggregateManager.AggregateManagerReply<AggregateManager3.VersionInfo> reply =
                am3.getVersion(getConnection());

        if (!reply.getGeniResponseCode().isSuccess()) {
            throw new JFedException("GetVersion Call not successful", reply);
        }

        return reply.getRawResult();
    }

    @Override
    public String listResources(List<AnyCredential> userCredentials, boolean available, String type, String version) throws JFedException {
        AbstractGeniAggregateManager.AggregateManagerReply<String> reply = am3.listResources(
                getConnection(),
                filterCredentials(userCredentials),
                type, version,
                available, true/*compressed*/,
                addCredentialExtraOptions(null)/*extraOptions*/);
        lastAmReply = reply;

        if (!reply.getGeniResponseCode().isSuccess()) {
            throw new JFedException("ListResources Call not successful", reply);
        }

        return reply.getValue();
    }

    public boolean hasSeperateAllocateAndProvision() {
        return true;
    }

    public static class AllocateReplyAmv3 implements AllocateReply {
        private final String manifest;
        private final List<GeniUrn> sliverUrns;

        //dirty trick to prevent erasure clash
        private AllocateReplyAmv3(List<AggregateManager3.SliverInfo> sliverInfos, String manifest) {
            this.manifest = manifest;
            this.sliverUrns = new ArrayList<GeniUrn>();
            if (sliverInfos != null)
                for (AggregateManager3.SliverInfo si : sliverInfos) {
                    GeniUrn sliverUrn = GeniUrn.parse(si.getSliverUrn());
                    if (sliverUrn != null)
                        sliverUrns.add(sliverUrn);
                }
        }

        public AllocateReplyAmv3(String manifest, List<GeniUrn> sliverUrns) {
            this.manifest = manifest;
            this.sliverUrns = sliverUrns;
        }

        @Override
        public String getManifest() {
            return manifest;
        }

        @Override
        public List<GeniUrn> getSliverUrns() {
            return null;
        }
    }
    public AllocateReply allocate(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, Date requestedExpirationDate)
            throws JFedException {
        String endTime = requestedExpirationDate == null ? null :
                RFC3339Util.dateToRFC3339String(requestedExpirationDate, true, true, true); //most compatible date format

        AbstractGeniAggregateManager.AggregateManagerReply<AggregateManager3.AllocateAndProvisionInfo> replyAllocate =
                am3.allocate(
                        getConnection(),
                        sliceUrn.getValue(), filterCredentials(sliceCredentials),
                        rspec,
                        endTime, addCredentialExtraOptions(null)/*extraOptions*/);
        lastAmReply = replyAllocate;
        if (!replyAllocate.getGeniResponseCode().isSuccess()) {
            if (AbstractGeniAggregateManager.isNotEnoughFreeResources(replyAllocate))
                throw new NotEnoughFreeResourcesException(replyAllocate);

            if (StitchingDirector.isAnyNotSupported(replyAllocate))
                throw new AnyVlanUnsupportedException(replyAllocate);

            if (StitchingDirector.isUnavailableVlan(replyAllocate))
                throw new VlanUnavailableException(StitchingDirector.getUnavailableVlan(replyAllocate), replyAllocate);

            throw new JFedException("Allocate Call not successful", replyAllocate);
        }
        return new AllocateReplyAmv3(
                replyAllocate != null && replyAllocate.getValue() != null ? replyAllocate.getValue().getSliverInfo() : null,
                replyAllocate.getValue().getRspec()
        );
    }

    public AllocateReply allocate(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, Date startDate, Date endDate)
            throws JFedException {
        String startTime = (startDate == null) ? null :
                RFC3339Util.dateToRFC3339String(startDate, true, true, true); //most compatible date format
        String endTime = (endDate == null) ? null :
                RFC3339Util.dateToRFC3339String(endDate, true, true, true); //most compatible date format

        Map<String, Object> extraOptionsStartTime = Collections.singletonMap("geni_start_time", (Object) startTime);

        AbstractGeniAggregateManager.AggregateManagerReply<AggregateManager3.AllocateAndProvisionInfo> replyAllocate =
                am3.allocate(
                        getConnection(),
                        sliceUrn.getValue(), filterCredentials(sliceCredentials),
                        rspec,
                        endTime, addCredentialExtraOptionsToMap(extraOptionsStartTime)/*extraOptions*/);
        lastAmReply = replyAllocate;
        if (!replyAllocate.getGeniResponseCode().isSuccess()) {
            if (AbstractGeniAggregateManager.isNotEnoughFreeResources(replyAllocate))
                throw new NotEnoughFreeResourcesException(replyAllocate);

            if (StitchingDirector.isAnyNotSupported(replyAllocate))
                throw new AnyVlanUnsupportedException(replyAllocate);

            if (StitchingDirector.isUnavailableVlan(replyAllocate))
                throw new VlanUnavailableException(StitchingDirector.getUnavailableVlan(replyAllocate), replyAllocate);

            throw new JFedException("Allocate Call not successful", replyAllocate);
        }
        return new AllocateReplyAmv3(
                replyAllocate != null && replyAllocate.getValue() != null ? replyAllocate.getValue().getSliverInfo() : null,
                replyAllocate.getValue().getRspec()
        );
    }


    public String provisionAndStart(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<UserSpec> users, Date requestedExpirationDate)
            throws JFedException {
        String endTime = requestedExpirationDate == null ? null :
                RFC3339Util.dateToRFC3339String(requestedExpirationDate, true, true, true); //most compatible date format

        List<String> urns = new ArrayList<String>();
        urns.add(sliceUrn.getValue());
        AbstractGeniAggregateManager.AggregateManagerReply<AggregateManager3.AllocateAndProvisionInfo> replyProvision =
                am3.provision(
                        getConnection(),
                        urns,
                        filterCredentials(sliceCredentials),
                        "geni", "3",
                        null/*besteffort*/,
                        endTime, users, addCredentialExtraOptions(null)/*extraOptions*/);
        lastAmReply = replyProvision;
        if (!replyProvision.getGeniResponseCode().isSuccess()) {

            if (AbstractGeniAggregateManager.isNotEnoughFreeResources(replyProvision)) //doesn't normally occur at Provision!
                throw new NotEnoughFreeResourcesException(replyProvision);

            if (StitchingDirector.isAnyNotSupported(replyProvision))
                throw new AnyVlanUnsupportedException(replyProvision);

            if (StitchingDirector.isUnavailableVlan(replyProvision))
                throw new VlanUnavailableException(StitchingDirector.getUnavailableVlan(replyProvision), replyProvision);

            throw new JFedException("Provision Call not successful", replyProvision);
            //            return null;
        }

        AbstractGeniAggregateManager.AggregateManagerReply<List<AggregateManager3.SliverInfo>> replyPerfOpAction =
                am3.performOperationalAction(
                        getConnection(),
                        urns,
                        filterCredentials(sliceCredentials),
                        "geni_start", null/*besteffort*/,
                        addCredentialExtraOptions(null)/*extraOptions*/);
        //        lastAmReply = replyPerfOpAction;
        lastAmReply = replyProvision; //SET LAST REPLY TO PROVISION REPLY!
        if (!replyPerfOpAction.getGeniResponseCode().isSuccess()) {
            if (AbstractGeniAggregateManager.isNotEnoughFreeResources(replyPerfOpAction)) //doesn't normally occur at PerformOperationAction!
                throw new NotEnoughFreeResourcesException(replyPerfOpAction);

            throw new JFedException("PerformOperationAction Call not successful", replyPerfOpAction);
            //            return null;
        }

        //check first sliver
        if (requestedExpirationDate != null &&
                replyProvision != null &&
                replyProvision.getGeniResponseCode().isSuccess() &&
                replyProvision.getValue() != null &&
                replyProvision.getValue().getSliverInfo() != null &&
                !replyProvision.getValue().getSliverInfo().isEmpty()) {
            try {
                Date actualExpiresDate = replyProvision.getValue().getSliverInfo().get(0).getExpiresDate();
                if (actualExpiresDate.after(requestedExpirationDate)) {
                    LOG.info("We requested a geni_end_time in the Provision call, but the AMv3 AM returned a later date! To try fix this, we execute a Renew.");

                    //but we wait first, since it will always be busy if you try immediately.
                    try {
                        Thread.sleep(30 * 1000L);
                    } catch (InterruptedException e) { /*ignore, doesn't matter if sleep worked or not*/ }

                    try {
                        renewSlivers(sliceUrn, null/*sliverUrns*/, filterCredentials(sliceCredentials), requestedExpirationDate);
                    } catch (JFedException ex) {
                        LOG.error("Error when trying to define expiration date for newly created sliver", ex);
                    }
                }
            } catch (ParseException e) {
                LOG.warn("Error parsing Provision reply geni_expires date \""+replyProvision.getValue().getSliverInfo().get(0).getExpires()+"\"");
            }
            lastAmReply = replyProvision; //SET LAST REPLY TO PROVISION REPLY!
        }

        return replyProvision.getValue().getRspec();
    }

    @Override
    public String createSliver(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, List<UserSpec> users, Date requestedExpirationDate)
            throws JFedException {
        allocate(sliceUrn, sliceCredentials, rspec, requestedExpirationDate);
        return provisionAndStart(sliceUrn, sliceCredentials, users, requestedExpirationDate);
    }

    @Override
    public void deleteSliver(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        List<String> urns = new ArrayList<String>();
        urns.add(sliceUrn.getValue());

        AbstractGeniAggregateManager.AggregateManagerReply<List<AggregateManager3.SliverInfo>> reply = am3.delete(
                getConnection(),
                urns,
                filterCredentials(sliceCredentials),
                null/*best effort*/,
                addCredentialExtraOptions(null)/*extraOptions*/);
        lastAmReply = reply;

        if (!reply.getGeniResponseCode().isSuccess()) {
            if (reply.getGeniResponseCode().equals(GeniAMResponseCode.GENIRESPONSE_SEARCHFAILED))
                return; //not an error, the sliver was already deleted, so all is ok.

            throw new JFedException("Delete Call not successful", reply);
            //            return null;
        }
    }

//    private SliverStatusDetails mergeSliverStatuses(SliverStatusDetails sliverStatus1, SliverStatusDetails sliverStatus2) {
//        if (sliverStatus1.equals(SliverStatusDetails.FAIL) || sliverStatus2.equals(SliverStatusDetails.FAIL)) {
//            List<String> errors = new ArrayList<String>();
//            errors.addAll(sliverStatus1.getErrors());
//            errors.addAll(sliverStatus2.getErrors());
//            return SliverStatusDetails.fail(errors);
//        }
//        if (sliverStatus1.getStatus().equals(SliverStatusDetails.SliverStatus.UNKNOWN) || sliverStatus2.getStatus().equals(SliverStatusDetails.SliverStatus.UNKNOWN))
//            return SliverStatusDetails.UNKNOWN;
//        if (sliverStatus1.getStatus().equals(SliverStatusDetails.SliverStatus.UNALLOCATED) || sliverStatus2.getStatus().equals(SliverStatusDetails.SliverStatus.UNALLOCATED))
//            return SliverStatusDetails.UNALLOCATED;
//        if (sliverStatus1.getStatus().equals(SliverStatusDetails.SliverStatus.UNINITIALISED) || sliverStatus2.getStatus().equals(SliverStatusDetails.SliverStatus.UNINITIALISED))
//            return SliverStatusDetails.UNINITIALISED;
//        if (sliverStatus1.getStatus().equals(SliverStatusDetails.SliverStatus.CHANGING) || sliverStatus2.getStatus().equals(SliverStatusDetails.SliverStatus.CHANGING))
//            return SliverStatusDetails.CHANGING;
//        if (sliverStatus1.getStatus().equals(SliverStatusDetails.SliverStatus.READY) && sliverStatus2.getStatus().equals(SliverStatusDetails.SliverStatus.READY))
//            return SliverStatusDetails.READY;
//        LOG.warn("unhandled mergeSliverStatuses case: " + sliverStatus1 + " merge with " + sliverStatus2 + "  (fix this! currently fallback to first of the 2)");
//        return sliverStatus1;
//    }

    @Override
    public StatusDetails status(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        LOG.debug("Calling AMv3Wrapper status with "+sliceCredentials.size()+" credentials");

        List<String> urns = new ArrayList<String>();
        urns.add(sliceUrn.getValue());

        AbstractGeniAggregateManager.AggregateManagerReply<AggregateManager3.StatusInfo> reply = am3.status(
                getConnection(),
                urns,
                filterCredentials(sliceCredentials),
                null/*best effort*/,
                addCredentialExtraOptions(null)/*extraOptions*/);
        lastAmReply = reply;

        AggregateManager3.StatusInfo status = reply.getValue();
        if ((!reply.getGeniResponseCode().isSuccess()) || status == null) {
            throw new JFedException("Status Call not successful", reply);
        }
        if (status.getSliverInfo().isEmpty())
            return new StatusDetails(StatusDetails.SliverStatus.UNALLOCATED); //no slivers


        StatusDetails res = null;
        for (AggregateManager3.SliverInfo sliverInfo : status.getSliverInfo()) {
            StatusDetails.SliverStatus s = getGlobalSliverStatus(sliverInfo);

            StatusDetails sliverStatusDetails = new StatusDetails(s, GeniUrn.parse(sliverInfo.getSliverUrn()));

            if (res != null) {
                res.add(sliverStatusDetails);
            } else {
                res = sliverStatusDetails;
            }

            if (sliverInfo.getExpires() != null)
                try {
                    res.addKnownExpires(sliverInfo.getExpiresDate());
                } catch (ParseException e) {
                    //ignore parse error
                    LOG.debug("Error parsing date in sliver_info "+sliverInfo.getExpires(), e);
                }

            if (sliverInfo.getError() != null && !sliverInfo.getError().isEmpty())
                res.addError(new ErrorDetails(sliverInfo.getError(), sliverInfo.getSliverUrn()));
        }

        return res;
    }

    @Override
    public String describe(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        List<String> urns = new ArrayList<String>();
        urns.add(sliceUrn.getValue());

        String type = "geni";
        String version = "3";
        if (amAuthority != null && amAuthority.getType() != null && amAuthority.getType().equalsIgnoreCase("FOAM"))
            type = "GENI"; //see FEDIBBTDEV-1392: workaround for bug at al2s (it should ignore case but doesn't)

        AbstractGeniAggregateManager.AggregateManagerReply<AggregateManager3.ManifestInfo> reply = am3.describe(
                getConnection(),
                urns,
                filterCredentials(sliceCredentials),
                type, version, true/*compressed*/,
                addCredentialExtraOptions(null)/*extraOptions*/);
        lastAmReply = reply;

        if (!reply.getGeniResponseCode().isSuccess()) {
            throw new JFedException("ListResources Call not successful", reply);
            //            return null;
        }

        return reply.getValue().getManifestRspec();
    }

    @Override
    public void renewSlivers(GeniUrn sliceUrn, List<GeniUrn> sliverUrns, List<AnyCredential> sliceCredentials, Date newExpire) throws JFedException {
        if (sliceUrn == null && (sliverUrns == null || sliverUrns.isEmpty()))
            throw new IllegalArgumentException("renewSlivers was called with both sliceUrn == null and sliverUrn empty or null (sliverUrns=" + sliverUrns + ")");

        List<String> urns = new ArrayList<String>();
        if (sliceUrn != null)
            urns.add(sliceUrn.getValue());
        else
            for (GeniUrn urn : sliverUrns)
                urns.add(urn.getValue());

        AbstractGeniAggregateManager.AggregateManagerReply<List<AggregateManager3.SliverInfo>> reply = am3.renew(
                getConnection(),
                urns,
                filterCredentials(sliceCredentials),
                newExpire,
                null/*best effort*/,
                addCredentialExtraOptions(null)/*extraOptions*/);
        lastAmReply = reply;

        if (!reply.getGeniResponseCode().isSuccess()) {
            throw new JFedException("Renew Call not successful", reply);
            //            return null;
        }
    }

    @Override
    public boolean isExirationdateSetByCreateSliver() {
        return true;
    }


    @Override
    public boolean isShareLanSupported() {
        if (amAuthority.getType().equalsIgnoreCase("emulab") || amAuthority.getType().equalsIgnoreCase("protogeni") || amAuthority.getType().equalsIgnoreCase("instageni"))
            return true;
        else
            return false;
    }

    @Override
    public void shareLan(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String lanClientId, String newSharedLanName) throws JFedException {
        List<String> urns = new ArrayList<String>();
        urns.add(sliceUrn.getValue());
        Map<String, Object> extraOptions = new HashMap<String, Object>();
        extraOptions.put("geni_sharelan_lanname", lanClientId);
        extraOptions.put("geni_sharelan_token", newSharedLanName);
        AbstractGeniAggregateManager.AggregateManagerReply<List<AggregateManager3.SliverInfo>> reply =
                am3.performOperationalAction(getConnection(), urns, filterCredentials(sliceCredentials), "geni_sharelan"/*action*/, false/*best_effort*/,
                        addCredentialExtraOptionsToMap(extraOptions)/*extraOptions*/);
        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Share lan command failed: " + reply.getOutput(), reply);
    }

    @Override
    public void unshareLan(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String lanClientId, String sharedLanName) throws JFedException {
        List<String> urns = new ArrayList<String>();
        urns.add(sliceUrn.getValue());
        Map<String, Object> extraOptions = new HashMap<String, Object>();
        extraOptions.put("geni_sharelan_lanname", lanClientId);
        extraOptions.put("geni_sharelan_token", sharedLanName);
        AbstractGeniAggregateManager.AggregateManagerReply<List<AggregateManager3.SliverInfo>> reply =
                am3.performOperationalAction(getConnection(), urns, filterCredentials(sliceCredentials), "geni_unsharelan"/*action*/, false/*best_effort*/,
                        addCredentialExtraOptionsToMap(extraOptions)/*extraOptions*/);
        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Unshare lan command failed: " + reply.getOutput(), reply);
    }


    @Override
    public boolean isRestartSupported() {
        //exception for planetlab
        if(amAuthority.isType("planetlab"))
            return false;

        return true;
    }

    @Override
    public boolean isReloadOSSupported() {

        //exception for planetlab
        if(amAuthority.isType("planetlab"))
            return false;

        return true;
    }

    @Override
    public boolean isOpenConsoleSupported() {
        //exception for planetlab
        if(amAuthority.isType("planetlab"))
            return false;

        //TODO actually depends on emulab version
        return true;
    }

    @Override
    public void restart(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        List<String> urns = new ArrayList<String>();
        urns.add(sliceUrn.getValue());
        AbstractGeniAggregateManager.AggregateManagerReply<List<AggregateManager3.SliverInfo>> reply =
                am3.performOperationalAction(getConnection(), urns, filterCredentials(sliceCredentials), "geni_restart"/*action*/, false/*best_effort*/, addCredentialExtraOptions(null)/*extraOptions*/);
        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Restart node command failed: " + reply.getOutput(), reply);
    }

    @Override
    public void reloadOS(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        List<String> urns = new ArrayList<String>();
        urns.add(sliceUrn.getValue());
        AbstractGeniAggregateManager.AggregateManagerReply<List<AggregateManager3.SliverInfo>> reply =
                am3.performOperationalAction(getConnection(), urns, filterCredentials(sliceCredentials), "geni_reload"/*action*/, false/*best_effort*/, addCredentialExtraOptions(null)/*extraOptions*/);
        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Reload OS on node command failed: " + reply.getOutput(), reply);
    }

    @Override
    public URL openConsole(GeniUrn sliverUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        List<String> urns = new ArrayList<String>();
        urns.add(sliverUrn.getValue());
        AbstractGeniAggregateManager.AggregateManagerReply reply =
                am3.performOperationalAction(getConnection(), urns, filterCredentials(sliceCredentials), "geni_console_url"/*action*/, false/*best_effort*/, addCredentialExtraOptions(null)/*extraOptions*/);
        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Get Console URL on node command failed: " + reply.getOutput(), reply);
        Object val = reply.getRawValue();
        if (val != null && val instanceof String) {
            try {
                return new URL((String)val);
            } catch (MalformedURLException e) {
                LOG.error("POA geni_console_url returned invalid URL: \""+val+"\"", e);
                return null;
            }
        }
        LOG.error("POA geni_console_url returned invalid URL: "+val+"");
        return null;
    }


    @Override
    public boolean isEditSshKeysSupported() {
        return true;
    }

    @Override
    public void editSshKeys(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, Collection<UserSpec> userSpecs) throws JFedException {
        List<String> urns = new ArrayList<String>();
        urns.add(sliceUrn.getValue());
        Map<String, Object> extraOptions = new HashMap<String, Object>();
        Vector userVect = new Vector();
        for (UserSpec u : userSpecs)
            userVect.add(u.getAsHashtable());
        extraOptions.put("geni_users", userVect);
        AbstractGeniAggregateManager.AggregateManagerReply<List<AggregateManager3.SliverInfo>> reply =
                am3.performOperationalAction(getConnection(), urns, filterCredentials(sliceCredentials), "geni_update_users"/*action*/, false/*best_effort*/, addCredentialExtraOptionsToMap(extraOptions)/*extraOptions*/);
        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Edit SSH keys command failed: " + reply.getOutput(), reply);
    }


    @Override
    public boolean areUserDiskImageFunctionsSupported() {
        return false;
    }

    @Override
    public ImageInfo createImage(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, GeniUrn componentUrn, String imageName, boolean global, boolean updatePrepare) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }



    @Override
    public void setSpeaksFor(@Nullable GeniUrn speakingForUrn) {
        super.setSpeaksFor(speakingForUrn);

        //NOTE: speaksFor credential itself should always be supplied as part of argument list of credentials each option takes
        if (speakingForUrn != null)
            setExtraOptionsForCallsWithCredential(Collections.singletonMap("geni_speaking_for", speakingForUrn.getValue()));
    }
}
