package be.iminds.ilabt.jfed.lowlevel.userloginmodel;

import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.LoggerFactory;

/**
 * UserLoginModel stores all preferences related to the user login:
 * - SSL key and certificate file
 * <p/>
 * it also stores the derived data, or allows overriding it:
 * - authority to use
 * - info about authority
 * (stores urn seperately, for cases where authority is not known, but urn is)
 * - username
 * - user urn
 * <p/>
 * - clearinghouse to use
 * It does not store the password, but it does store the unencrypted private key
 * <p/>
 * It is an implementation of GeniUser that is used as the underlying model for the UserLoginPanel
 * <p/>
 * Note that UserLoginModel can be in invalid states, where this is not expected users of the GeniUser interface
 */
public class KeyCertWithManualInfoUserLoginModel extends KeyCertUserLoginModel {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(KeyCertWithManualInfoUserLoginModel.class);
    private static final String PREFS_KEY_CERT_FILE_URI = "keyCertFileURI";
    private static final String PREFS_USER_AUTH_URN = "userauthurn";
    private static final String PREFS_USER_URN = "userurn";

    public KeyCertWithManualInfoUserLoginModel(AuthorityListModel authorityListModel, AuthorityFinder authorityFinder, UserLoginModelManager userLoginModelManager) {
        super(authorityListModel, authorityFinder, userLoginModelManager);
    }

    @Override
    protected void processPemContent() throws InvalidLoginException {
        processBasicPemContent();
    }

    public void setAuthority(SfaAuthority auth) {
        this.authority = auth;
        if (authority != null)
            this.authorityUrn = auth.getUrnString();
        else
            this.authorityUrn = null;
    }

    public void setUserUrn(String urn) {
        if (urn == null) {
            this.userUrn = null;
            return;
        }
        GeniUrn parsedUrn = null;
        try {
            parsedUrn = new GeniUrn(urn);
            this.userUrn = parsedUrn;
        } catch (GeniUrn.GeniUrnParseException e) {
            LOG.warn("provided urn is NOT a valid urn");
            this.userUrn = null;
        }
    }

    public boolean isValidUserUrn() {
        if (userUrn == null) return false;
        if (!userUrn.getEncodedResourceType().equals("user")) return false;
        return true;
    }

    public boolean deriveAuthorityFromUrn() {
        if (userUrn == null) return false;

        authority = authorityFinder.findByAnyUrn(authorityUrn, AuthorityFinder.Purpose.FIND_USERAUTH);
        if (authority != null) {
            authorityUrn = authority.getUrn().toString();
            return true;
        } else {
            authorityUrn = null;
            return false;
        }
    }

    public boolean authorityMatchesUrn() {
        if (authority == null) return false;
        if (userUrn == null) return false;
        String userAuthPart = userUrn.getEncodedTopLevelAuthority();

        return authority.getNameForUrn().equals(userAuthPart);
    }

    @Override
    public void save(UserLoginModelManager.UserLoginModelPreferences prefs) {
        super.save(prefs);
        prefs.put(PREFS_KEY_CERT_FILE_URI, keyCertFile.toURI().toString());

        if (authorityUrn != null)
            prefs.put(PREFS_USER_AUTH_URN, authorityUrn);
        else
            prefs.remove(PREFS_USER_AUTH_URN);

        if (userUrn != null)
            prefs.put(PREFS_USER_URN, userUrn.getValue());
        else
            prefs.remove(PREFS_USER_URN);
    }

    @Override
    public void load(UserLoginModelManager.UserLoginModelPreferences prefs) throws InvalidLoginException {
        super.load(prefs);

        authority = null;
        authorityUrn = prefs.get(PREFS_USER_AUTH_URN, null);
        userUrn = GeniUrn.parse(prefs.get(PREFS_USER_URN, null));

        if (authorityUrn != null)
            authority = authorityFinder.findByUrn(authorityUrn, AuthorityFinder.Purpose.FIND_USERAUTH);
    }
}
