package be.iminds.ilabt.jfed.rspec.parser.extensions;

import be.iminds.ilabt.jfed.rspec.model.bonfire.DiskImage;
import be.iminds.ilabt.jfed.rspec.model.bonfire.Lan;
import be.iminds.ilabt.jfed.rspec.parser.EventListExtraXml;
import be.iminds.ilabt.jfed.rspec.parser.ExtraXml;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.util.TextUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;
import java.util.ArrayList;
import java.util.List;

import static be.iminds.ilabt.jfed.rspec.parser.RspecXmlConstants.*;

/**
 * BonfireParser
 */
public class BonfireParser {

    public static DiskImage parseDiskImage(StartElement startElement, XMLEventReader eventReader) throws XMLStreamException, RspecParseException {
        assert startElement.getName().equals(Q_BONFIRE_DISK_IMAGE);

        Attribute attComManId = startElement.getAttributeByName(Q_ATT_COM_MAN_ID);
        Attribute attComId = startElement.getAttributeByName(Q_ATT_COM_ID);
        Attribute attComName = startElement.getAttributeByName(Q_ATT_COM_NAME);
        Attribute attExclusive = startElement.getAttributeByName(Q_ATT_NODE_EXCLUSIVE);

        if (attComManId == null)
            LOG.error("Expected attribute component_manager_id in disk_image, but got none @ {}", startElement.getLocation());
        if (attComId == null)
            LOG.error("Expected attribute component_id in disk_image, but got none @ {}", startElement.getLocation());
        if (attComName == null)
            LOG.error("Expected attribute component_id in component_name, but got none @ {}", startElement.getLocation());
        Boolean exclusive = attExclusive == null ? null : TextUtil.stringToBoolean(attExclusive.getValue());

        String type = null;
        String desc = null;
        List<ExtraXml> extraXmls = new ArrayList<ExtraXml>();
        boolean nested = false;
        while (eventReader.hasNext()) {
            XMLEvent e = eventReader.nextEvent();

            if (e.isStartElement()) {
                StartElement se = e.asStartElement();
                if (nested)
                    throw new RspecParseException(e.getLocation(), "Unexpected nesting is causing parse failure in bonfire disk_image for " + se.getName());
                nested = true;

                boolean handled = false;
                if (se.getName().equals(Q_OCCI_TYPE)) {
                    e = eventReader.nextEvent();

                    if (!e.isCharacters())
                        throw new RspecParseException(e.getLocation(), "Expected characters in " + se.getName());
                    type = e.asCharacters().getData();
                    handled = true;
                }
                if (se.getName().equals(Q_OCCI_DESCRIPTION)) {
                    e = eventReader.nextEvent();

                    if (e.isCharacters()) {
                        desc = e.asCharacters().getData();
                    }
                    //else, it will be handled later in this loop if it is the end element
                    handled = true;
                }

                if (!handled) {
                    EventListExtraXml extraXml = new EventListExtraXml(se, eventReader);
                    extraXmls.add(extraXml);
                    nested = false;
                    continue;
                }
            }

            if (e.isEndElement()) {
                nested = false;
            }

            if (e.isEndElement() && e.asEndElement().getName().equals(Q_BONFIRE_DISK_IMAGE)) {
                if (type == null)
                    LOG.warn("Expected element occi:type in disk_image, but got none @ {}", startElement.getLocation());

                return new DiskImage(type, desc,
                        attComId != null ?  attComId.getValue() : null,
                        attComManId != null ? attComManId.getValue() : null,
                        attComName != null ? attComName.getValue() : null,
                        exclusive, extraXmls);
            }
        }

        throw new RspecParseException(null, "Did not encounter end element for " + Q_JFEDSSH_USER_SSH_KEYS);
    }

    private static final Logger LOG = LoggerFactory.getLogger(BonfireParser.class);

    public static Lan parseLan(StartElement startElement, XMLEventReader eventReader) throws XMLStreamException, RspecParseException {
        assert startElement.getName().equals(Q_BONFIRE_LAN);

        Attribute attComManId = startElement.getAttributeByName(Q_ATT_COM_MAN_ID);
        Attribute attComId = startElement.getAttributeByName(Q_ATT_COM_ID);
        Attribute attComName = startElement.getAttributeByName(Q_ATT_COM_NAME);
        Attribute attExclusive = startElement.getAttributeByName(Q_ATT_NODE_EXCLUSIVE);

        if (attComManId == null)
             throw new RspecParseException(startElement.getLocation(), "Expected attribute component_manager_id in lan, but got none");
//        if (attComId == null)
//            throw new RspecParseException(startElement.getLocation(), "Expected attribute component_id in lan, but got none");
//        if (attComName == null)
//            throw new RspecParseException(startElement.getLocation(), "Expected attribute component_name in lan, but got none");

        if (attComId == null && attComName == null)
            throw new RspecParseException(startElement.getLocation(), "Expected attribute component_name or component_id in lan, but got none");

        Boolean exclusive = attExclusive == null ? null : TextUtil.stringToBoolean(attExclusive.getValue());

        boolean nested = false;
        List<ExtraXml> extraXmls = new ArrayList<ExtraXml>();
        while (eventReader.hasNext()) {
            XMLEvent e = eventReader.nextEvent();

            if (e.isStartElement()) {
                StartElement se = e.asStartElement();

                if (nested)
                    throw new RspecParseException(e.getLocation(), "Unexpected nesting is causing parse failure in bonfire lan, for " + se.getName());
                boolean handled = false;
                nested = true;

                //at the moment, we don't parse occi:vlan or occi:netmask

                if (!handled) {
                    EventListExtraXml extraXml = new EventListExtraXml(se, eventReader);
                    extraXmls.add(extraXml);
                    nested = false;
                    continue;
                }
            }

            if (e.isEndElement())
                nested = false;

            if(e.isEndElement() && e.asEndElement().getName().equals(Q_BONFIRE_LAN)) {
                return new Lan(
                        attComId == null ? null : attComId.getValue(),
                        attComManId == null ? null : attComManId.getValue(),
                        attComName == null ? null : attComName.getValue(),
                        exclusive, extraXmls);
            }
        }

        throw new RspecParseException(null, "Did not encounter end element for " + Q_JFEDSSH_USER_SSH_KEYS);
    }
}
